import datetime
import numpy as np
from scipy import integrate
import numpy.testing as npt
import varwg
from varwg.time_series_analysis import _kde as kde
from varwg import times


class Test(npt.TestCase):
    def setUp(self):
        n_years = 2
        # construct n_years of hourly data
        self.dtimes = np.array(
            [
                datetime.datetime(2000, 1, 1) + t * datetime.timedelta(hours=1)
                for t in np.arange(n_years * 365 * 24)
            ]
        )
        x = np.arange(len(self.dtimes))
        # this is roughly a daily cycle within a yearly cycle
        # the strength of the daily cycle is dependent on the doy
        self.noise = varwg.rng.normal(size=len(x))
        self.data = (
            15 * -np.cos(x * 2 * np.pi / (365 * 24))
            + 5
            * (1 + np.cos(x * 2 * np.pi / (365 * 24)))
            * -np.cos(x * 2 * np.pi / 24)
            + 1.5 * self.noise
        )
        self.circ = circ = 1.5
        self.dists_exp = [
            0.0,
            ((6.0 / 24) ** 2 + (circ * 6 / 24.0) ** 2) ** 0.5,
            ((12.0 / 24) ** 2 + (circ * 12 / 24.0) ** 2) ** 0.5,
            ((18.0 / 24) ** 2 + (circ * 6 / 24.0) ** 2) ** 0.5,
            1.0,
            ((30.0 / 24) ** 2 + (circ * 6 / 24.0) ** 2) ** 0.5,
            ((36.0 / 24) ** 2 + (circ * 12 / 24.0) ** 2) ** 0.5,
            ((42.0 / 24) ** 2 + (circ * 6 / 24.0) ** 2) ** 0.5,
        ]
        doys = times.datetime2doy(self.dtimes[: 2 * 24 : 6])
        n_dist = len(self.dists_exp)
        self.dists_exp2 = [
            [
                kde.doyhour_distance(
                    doys[i], doys[j], doy_width=1, circ=self.circ
                )
                for i in range(n_dist)
            ]
            for j in range(n_dist)
        ]

    def tearDown(self):
        pass

    def test_kernel_integration(self):
        varwg.reseed(0)
        sample = varwg.rng.normal(size=5000)
        q1, q2 = 0.25, 0.75
        x0, x1 = map(lambda q: np.percentile(sample, 100 * q), (q1, q2))
        kernel_width = kde.optimal_kernel_width(sample)

        def density_func(x):
            return kde.kernel_density(kernel_width, sample, eval_points=x)

        integral, abserr = integrate.quad(density_func, x0, x1, limit=200)

        # import matplotlib.pyplot as plt
        # xx = np.linspace(x0, x1, 500)
        # densities = density_func(xx)
        # xx2 = np.linspace(sample.min(), sample.max(), 100)
        # densities2 = density_func(xx2)
        # integrals2 = integrate.cumtrapz(y=densities2, x=xx2, initial=0)
        # sample_sorted = sample.copy()
        # sample_sorted.sort()
        # sample_ranks = (np.arange(1., len(sample) + 1.) - .5) / len(sample)
        # plt.plot(xx2, densities2, label="kde all")
        # plt.plot(xx, densities, label="kde test")
        # plt.plot(xx2, integrals2, label="kde cdf all")
        # plt.plot(sample_sorted, sample_ranks, label="sample cdf")
        # plt.scatter(sample, np.zeros_like(sample), label="sample")
        # plt.legend(loc="best")
        # plt.show()

        npt.assert_almost_equal(integral, q2 - q1, decimal=2)

    def test_log_kernel_integration(self):
        varwg.reseed(0)
        sample = np.exp(np.linspace(0, 1, 1000))
        kernel_width = kde.optimal_kernel_width(np.log(sample))
        q1, q2 = 1e-6, 1 - 1e-6
        x0, x1 = map(lambda q: np.percentile(sample, 100 * q), (q1, q2))

        def density_func(x):
            density = kde.kernel_density(
                kernel_width, np.log(sample), eval_points=np.log(x)
            )
            return density / x

        integral = integrate.quad(density_func, x0, x1)[0]
        npt.assert_almost_equal(integral, q2 - q1, decimal=3)

    def test_distance_array_sparse(self):
        """Do we get the same as with distance_array?"""
        x_vec = np.arange(10)
        dists_exp = x_vec[None, :] - x_vec[:, None]
        mask = dists_exp < 0
        dists_exp[mask] = 0
        dists_act = kde.distance_array_sparse(x_vec, x_vec, mask)
        npt.assert_almost_equal(dists_exp, dists_act.toarray())
        # kde.distance_array_sparse.clear_cache()

    def test_apply_sparse_kernel(self):
        """Does sparse output the same as dense?"""
        data = np.arange(50)
        width = 5.0
        mask = np.full((len(data), len(data)), True, dtype=bool)
        mask_mask = (slice(None, None, 2), slice(None, None, 4))
        mask_sparse = np.copy(mask)
        mask_sparse[mask_mask] = False
        kde.apply_kernel.clear_cache()
        dens_exp = kde.apply_kernel(width, data)
        dens_exp_sparse = np.copy(dens_exp)
        dens_exp_sparse[~mask_sparse] = 0.0
        for kernel in (kde.gaussian_kernel, kde.gaussian_kernel_ne):
            # let's start with all unmasked points
            dens_act = kde.apply_sparse_kernel(kernel, width, data, mask)
            npt.assert_almost_equal(dens_exp, dens_act.toarray())
            kde.apply_sparse_kernel.clear_cache()
            # some masked points
            dens_act = kde.apply_sparse_kernel(
                kernel, width, data, mask_sparse
            )
            npt.assert_almost_equal(dens_exp_sparse, dens_act.toarray())
            kde.apply_sparse_kernel.clear_cache()

    def test_doyhour_distance_dt(self):
        dtimes = self.dtimes[: 2 * 24 : 6]
        from_time = self.dtimes[0]
        dists = kde.doyhour_distance_dt(from_time, dtimes, circ=self.circ)
        npt.assert_almost_equal(dists, self.dists_exp)

    def test_doyhour_distance(self):
        doys = kde.times.datetime2doy(self.dtimes[: 2 * 24 : 6])
        from_time = doys[0]
        dists = kde.doyhour_distance(
            from_time, doys, doy_width=1, circ=self.circ
        )
        npt.assert_almost_equal(dists, self.dists_exp)
        dists = kde.doyhour_distance(doys, doys, doy_width=1, circ=self.circ)
        npt.assert_almost_equal(dists, self.dists_exp2)
        dist = kde.doyhour_distance(1.0, 365.0, 1.0, 1.0)
        npt.assert_almost_equal(dist, 1.0)
        dist = kde.doyhour_distance(1.0, 365.0 + 23.0 / 24, 1.0, 1.0)
        npt.assert_almost_equal(dist, np.sqrt(2 * (1.0 / 24) ** 2))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    # import warnings
    # warnings.simplefilter("error", RuntimeWarning)
    npt.run_module_suite()



"""Just some functions I find myself writing or searching for again and
again..."""

# The imports are not all done here.  Importing this module to use a few
# specific functions would mean pulling a lot of other imports in that are not
# needed.  As importing some of the modules can be quite costly speed-wise,
# the clarity of having the import statements at the beginning is consciously
# compromised.
from collections import UserDict
import contextlib
import datetime
import inspect
import itertools
import numbers
import os
import re
import sys
import warnings


try:
    # should make multiprocessing less stressfull
    import dill as pickle
except ImportError:
    import pickle

import hashlib
import functools
import random
import numpy as np
import numpy.testing as npt

try:
    from multiprocessing import cpu_count
    import numexpr as ne

    ne.set_num_threads(min(64, cpu_count()))
    NE = True
except ImportError:
    NE = False
from scipy import optimize, stats
from scipy.stats import rankdata
import scipy
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.patches as mpatches

PY2 = sys.version_info.major == 2

# Generic


def asscalar(func):
    """Return the result as a scalar if it has len == 1."""

    @functools.wraps(func)
    def wrapped(*args, **kwds):
        result = np.atleast_1d(func(*args, **kwds))
        return result.item() if result.size == 1 else result

    return wrapped


def _build_arg_str(function, *func_args, **func_kwds):
    """Builds a string that completely describes the parameters in a function
    call. This helps to cache function results that have no side-effects."""
    # remember that default arguments arrive empty in **func_kwds.
    # build dictionary of argument-names to default argument values
    spec_arg_names = inspect.getfullargspec(function).args
    if spec_arg_names[0] == "self":
        spec_arg_names = spec_arg_names[1:]
    arg_dict = dict.fromkeys(spec_arg_names, None)
    spec_arg_values = inspect.getfullargspec(function).defaults
    # spec_arg_values apply to the last n spec_arg_names
    if spec_arg_names and spec_arg_values:
        for name, val in zip(spec_arg_names[::-1], spec_arg_values[::-1]):
            arg_dict[name] = val

    # change the default values according to the passed values
    for arg, arg_name in zip(func_args, spec_arg_names):
        arg_dict[arg_name] = arg
    arg_dict.update(func_kwds)

    return "__".join(
        "%s_%s" % (key, arg_dict[key]) for key in sorted(arg_dict.keys())
    )


def pickle_cache(filepath_template="%s.pkl", clear_cache=False, warn=True):
    """Use this as a function decorator to cache the result of a function as a
    pickle-file. The filename is determined by the arguments in the function.
    If the filename turns out to be too long, a hash of it is used."""

    def function_wrapper(function):
        @functools.wraps(function)
        def pickle_function(*args, **kwds):
            # set a filename based on the arguments passed
            # arg_dict_str = sanitize(_build_arg_str(function, *args, **kwds))
            arg_dict_str = _build_arg_str(function, *args, **kwds)
            if len(arg_dict_str) > 0:
                # if isinstance(filepath_template, pathlib.Path):
                #     filepath_template = str(filepath_template)
                filepath = filepath_template % arg_dict_str
            else:
                filepath = filepath_template

            def hash_(filepath):
                name_hash = hashlib.md5(
                    os.path.basename(filepath.encode())
                ).hexdigest()
                return os.path.join(os.path.dirname(filepath), name_hash)

            filepath_hash = hash_(filepath)
            if clear_cache:
                if os.path.exists(filepath):
                    os.remove(filepath)
                if os.path.exists(filepath_hash):
                    os.remove(filepath_hash)

            if not re.match(r".*py[2,3]\.pkl$", filepath):
                # keep different pickles for different python versions
                version_str = "_%s." % ("py2" if PY2 else "py3")
                filepath = version_str.join(filepath.rsplit("."))
                # name_parts = filepath.rsplit(".")
                # filepath = "".join((name_parts[0],
                #                     "_",
                #                     "py2" if PY2 else "py3",
                #                     name_parts[1]))

            def read(filepath):
                with open(filepath, "rb") as pi_file:
                    return pickle.load(pi_file)

            if os.path.exists(filepath) or os.path.exists(hash_(filepath)):
                try:
                    if warn:
                        warnings.warn(
                            "I am not executing %s, but restoring the "
                            % repr(function)
                            + "result from its last execution with the "
                            + "same parameters. \nThe result is stored "
                            + "here: %s" % filepath
                        )
                    return read(filepath)
                except IOError:
                    if os.path.exists(filepath_hash):
                        return read(filepath_hash)
                except UnicodeDecodeError:
                    # this is probably a py2/3 incompatibility problem
                    os.remove(filepath)

            result = function(*args, **kwds)

            def dump(filepath):
                with open(filepath, "wb") as pi_file:
                    pickle.dump(result, pi_file)

            try:
                dump(filepath)
            except OSError:
                dump(hash_(filepath))
            except IOError:
                # maybe the filename was too long
                dump(hash_(filepath))
            return result

        return pickle_function

    return function_wrapper


class ADict(UserDict):
    def __add__(self, other):
        # we need a copy to work with
        left_dict = dict(self)
        left_dict.update(other)
        # make sure we can do this operation also with the returned
        # object
        return ADict(left_dict)

    def __sub__(self, other):
        left_dict = dict(self)
        if isinstance(other, dict):
            del_keys = list(other.keys())
        elif isinstance(other, str):
            del_keys = (other,)
        else:
            del_keys = other
        for del_key in del_keys:
            del left_dict[del_key]
        return ADict(left_dict)


def cache(*names, **name_values):
    """Use as a decorator, to supply *names attributes that can be used as
    a cache. The attributes are set to their default/None during
    compile time. The wrapped function also has a 'clear_cache'-method
    to delete those variables.

    Parameters
    ----------
    *names : str

    """

    def wrapper(function):
        @functools.wraps(function)
        def cache_holder(*args, **kwds):
            return function(*args, **kwds)

        cache_holder._cache_names = names
        cache_holder._cache_name_values = name_values
        cache_holder.clear_cache = lambda: clear_def_cache(cache_holder)
        cache_holder.clear_cache()
        return cache_holder

    return wrapper


def clear_def_cache(function, cache_names=None, cache_name_values=None):
    """I often use a simplified function cache in the form of
    'function.attribute = value'.  This function helps cleaning it up,
    i.e. setting them to None.

    Parameters
    ----------
    function : object with settable attributes
    cache_names : sequence of str or None, optional
        if None, function should have an attribute called _cache_names with
        names of attributes that are cached.
    """
    if cache_names is None:
        cache_names = function._cache_names
    if cache_name_values is None:
        cache_name_values = function._cache_name_values
    for name in cache_names:
        setattr(function, name, None)
    for name, value in list(cache_name_values.items()):
        setattr(function, name, value)


def _plot_array_diff(name, array1, array2):
    array1, array2 = map(np.atleast_2d, (array1, array2))
    nrows1, ncols1 = array1.shape
    nrows2, ncols2 = array2.shape
    if nrows1 != nrows2:
        print(
            "Cannot plot with different number of variables"
            + f"({nrows1} != {nrows2})"
        )
        return None, None
    fig, axs = plt.subplots(
        nrows=nrows1,
        ncols=2,
        constrained_layout=True,
        sharex="col",
        sharey="row",
        width_ratios=(0.8, 0.2),
    )
    if nrows1 == 1:
        axs = np.atleast_2d(axs)
    for row_i, ax in enumerate(axs):
        ax[0].plot(array1[row_i], label="1")
        ax[0].plot(array2[row_i], label="2")
        if ncols1 == ncols2:
            ax[1].scatter(
                array1[row_i],
                array2[row_i],
                marker="o",
                edgecolor=(0, 0, 0, 0),
                facecolor=(1, 1, 1, 0.5),
            )
        else:
            ax[1].text(
                0.5,
                0.5,
                "Different number of records",
                horizontalalignment="center",
                verticalalignment="center",
            )
        ax[1].set_aspect("equal", "box")
    axs[0, 0].legend(loc="best")
    for ax in np.ravel(axs):
        ax.grid(True)
    fig.suptitle(name)
    return fig, axs


@cache("fig_axs")
def recursive_diff(
    name,
    obj1,
    obj2,
    *,
    ignore_types=None,
    plot=False,
    verbose=False,
    diff=None,
):
    """Show differences between two objects.

    For debugging purposes mostly. Recurses into sequences and instances of the current module.

    Parameters
    ----------
    name : str or None
        Name of the current object. If None, will be determined by str(obj1).
    obj1 : object
        First object to compare.
    obj2 : object
        Second object to compare.
    ignore_types : sequence or None, optional
        Ignore differences for objects of these types. callables are ignored by default.
    plot : bool, optional
        Plot 1- and 2-D np.ndarrays if they are different.
    verbose : bool, optional
        Be more talkative.

    Examples
    --------
    tba

    """
    if name is None:
        name = str(obj1)
    if ignore_types is None:
        ignore_types = (scipy.interpolate.interpolate.interp1d, np.vectorize)
    if diff is None:
        diff = {}
    r_kwds = dict(ignore_types=ignore_types, plot=plot, verbose=verbose)
    name = f"  {name}"
    if isinstance(obj1, ignore_types) or isinstance(obj2, ignore_types):
        if verbose > 1:
            print(f"{name}: ignoring type {type(obj1)}")
        return diff
    if any(
        itertools.chain(
            map(inspect.isfunction, (obj1, obj2)),
            map(inspect.isgenerator, (obj1, obj2)),
            # map(lambda x: hasattr(x, "__call__"), (obj1, obj2)),
        )
    ):
        if verbose > 1:
            print(f"{name}: ignoring callable {obj1}")
        return diff
    if plot and recursive_diff.fig_axs is None:
        recursive_diff.fig_axs = {}

    if type(obj1) != type(obj2):
        if verbose:
            print(f"{name}: {type(obj1)} != {type(obj2)}")
        diff[name.lstrip()] = obj1, obj2
        return diff
    if isinstance(obj1, np.ndarray):
        try:
            if isinstance(obj1[0], datetime.datetime):
                obj1 = np.array([np.datetime64(value) for value in obj1])
                obj2 = np.array([np.datetime64(value) for value in obj2])
                assert np.all(obj2 == obj1)
                if verbose > 1:
                    print(f"{name}: arrays are approximately equal")
                return False
            if obj1.shape == obj2.shape:
                npt.assert_almost_equal(obj1, obj2, verbose=True)
            else:
                if verbose:
                    print(
                        f"{name}: arrays have different shape "
                        + f"({obj1.shape} != {obj2.shape})"
                    )
                diff[name.lstrip()] = obj1, obj2
        except AssertionError as exc:
            if verbose:
                print(name)
                print(exc)
            if plot and obj1.dtype != bool and obj1.shape[0] < 20:
                recursive_diff.fig_axs[name.lstrip()] = _plot_array_diff(
                    name, obj1, obj2
                )
                diff[name.lstrip()] = obj1, obj2
            return diff
    elif isinstance(obj1, dict):
        # check keys on both
        keys1, keys2 = set(obj1.keys()), set(obj2.keys())
        if extra1 := keys1 - keys2:
            if verbose:
                print(f"{name}: {obj1} has extra keys: {extra1}")
            diff[f"{name.lstrip()}_obj1_extra_keys"] = (
                obj1,
                obj2,
            )
        if extra2 := keys2 - keys1:
            if verbose:
                print(f"{name}: {obj2} has extra keys: {extra2}")
            diff[f"{name.lstrip()}_obj2_extra_keys"] = (
                obj1,
                obj2,
            )
        diff_sub = {}
        for key in keys1 & keys2:
            if diff_element := recursive_diff(
                f"{key}", obj1[key], obj2[key], diff=None, **r_kwds
            ):
                if len(diff_element) == 1 and isinstance(diff_element, dict):
                    diff_element = diff_element[list(diff_element.keys())[0]]
                diff_sub[key] = diff_element
        if diff_sub:
            diff[name.lstrip()] = diff_sub
        return diff
    elif isinstance(obj1, (list, tuple, set)):
        diff_sub = []
        for elem_i, (elem1, elem2) in enumerate(zip(obj1, obj2)):
            if diff_element := recursive_diff(
                f"{elem_i}", elem1, elem2, diff=None, **r_kwds
            ):
                diff_sub += [diff_element]

        if diff_sub:
            diff[name.lstrip()] = diff_sub
        return diff
    elif (
        hasattr(obj1, "__module__")
        and obj1.__module__.split(".")[0] == __name__.split(".")[0]
    ):  # HACK! only consider differences arising from our own code.
        return recursive_diff(
            # f"{name}{obj1}",
            f"{obj1}",
            obj1.__dict__,
            obj2.__dict__,
            diff=diff,
            **r_kwds,
        )
    elif obj1 != obj2:
        if (
            isinstance(obj1, numbers.Number)
            and isinstance(obj2, numbers.Number)
            and np.all(np.isnan([obj1, obj2]))
        ):
            return False
        if verbose:
            print(f"{name}: {obj1} (self) != {obj2} (other)")
        return obj1, obj2
    return diff


def key_tree(dict_, level=0):
    for key, value in dict_.items():
        print(" " * level + key)
        if isinstance(value, dict):
            key_tree(value, level + 1)


# Filesystem


@contextlib.contextmanager
def chdir(dirname):
    """Temporarily change the working directory with a with-statement."""
    old_dir = os.path.abspath(os.path.curdir)
    if dirname:  # could be an empty string
        os.chdir(dirname)
    yield
    os.chdir(old_dir)


# Numeric


def kendalls_tau(x, y):
    """Kendall's Rank correlation coefficient. See Hartung p.599f

    Examples
    --------
    >>> A = [8, 6, 5, 3.5, 1, 2, 3.5, 7]
    >>> B = [6, 7.5, 4, 1, 2, 3, 5, 7.5]
    >>> kendalls_tau(A, B)
    0.5714285714285714
    """
    assert len(x) == len(y)
    x, y = np.asarray(x), np.asarray(y)
    n = len(x)
    # TODO: we have to be pessimistic when coming across equal values
    y_xrank_sorted = y[np.argsort(x)]
    y_ranks = rankdata(y_xrank_sorted)
    # i forcefully put in "<" instead of "<=" because i am annoyed that the
    # correlation between a variable and itself is not 1 when there are
    # equal values inside
    # comment on the comment: changed that back to do it "by the book"
    q_i = np.array(
        [np.sum(y_ranks[ii + 1 :] <= y_ranks[ii]) for ii in range(n)],
        dtype=float,
    )
    return 1 - 4 * np.sum(q_i) / (n * (n - 1))


def chi2_test(x, y, k=None, n_parameters=0):
    """Chi-square test for inequality.
    H0: x and y were sampled from the same distribution.

    Parameters
    ----------
    k : int
        Number of classes (bins)

    Returns
    -------
    p_value : float
    """
    n = len(x)
    if k is None:
        k = int(n**0.5)
        # k = n_parameters + 2
    observed, bins = np.histogram(x, k)[:2]
    expected = np.histogram(y, bins)[0]
    chi_test = np.sum((observed.astype(float) - expected) ** 2 / expected)
    # degrees of freedom:
    dof = k - n_parameters - 1
    print(chi_test, stats.chi2.ppf(0.95, dof))
    return stats.chisqprob(chi_test, dof)


def rel_ranks(values, method="average"):
    """Returns ranks of values in the range [0,1]."""
    if isinstance(values, int):
        N = values
        return (np.arange(N) + 0.5) / N
    return (rankdata(values, method) - 0.5) / len(values)


def val2ind(values, value):
    """Return the index of the nearest neighbor of value in values."""
    # the int-conversion is necessary.  without it, the index comes out as
    # 'numpy.int64' (on my machine), which causes "illegal subscript type"
    # errors when used as an index on netcdf-arrays.
    flat_index = int(np.argmin(np.abs(values - value)))
    if np.ndim(values) == 1:
        return flat_index
    else:
        return np.unravel_index(flat_index, values.shape)


def round_to_float(values, precision):
    """Round to nearest precision.

    >>> round_to_float([8, 12], 5.)
    array([ 10.,  10.])
    """
    values = np.asarray(values, dtype=float)
    rest = values % precision
    return np.where(
        rest > precision / 2.0,
        values + (precision - rest),
        values - rest,
    )


def fourier_approx(data, order=4, size=None, how="longest"):
    """Approximate data with a Fourier transform, using the order number of
    frequencies with the highest amplitudes.

    Parameters
    ----------
    data : 1-dim ndarray
    order : int, optional
        Number of frequencies to account for.
    size : int, optional
        Desired length of the output. If None, it will be the same as data.
    how : "longest" or "strongest", optional
    """
    if size is None:
        size = len(data)
    data_freq = np.fft.fft(data)
    if how == "longest":
        ii_below = list(range(order + 1, len(data)))
    elif how == "strongest":
        # find the order biggest amplitudes
        ii_below = np.argsort(np.abs(data_freq))[: len(data_freq) - order - 1]
    pars = np.copy(data_freq)
    pars[ii_below] = 0
    return np.fft.irfft(pars, size)


def periodic_pad(values):
    half = len(values) // 2
    return np.concatenate((values[-half:], values, values[:half]))


def interp_nonfin(
    values, times=None, max_interp=None, pad_periodic=False, mask=None
):
    """Remove nans from values by linear interpolation.

    Parameters
    ----------
    values : 2d array
    times : 1d array, optional
    max_interp : int
        Maximum number of subsequent nans to interpolate over.

    Examples
    --------
    >>> import numpy as np
    >>> a = np.array([0., np.nan, 1., np.nan, np.nan, 4.])
    >>> interp_nan(a)
    array([ 0. ,  0.5,  1. ,  2. ,  3. ,  4. ])
    >>> interp_nan(a, max_interp=1)
    array([ 0. ,  0.5,  1. ,  nan,  nan,  4. ])
    >>> a = np.arange(6, dtype=float).reshape((2, 3))
    >>> a[0, 1] = np.nan
    >>> interp_nan(a)
    array([[ 0.,  1.,  2.],
           [ 3.,  4.,  5.]])
    """
    if pad_periodic:
        half = len(values) // 2
        values = periodic_pad(values)
    values = np.atleast_2d(np.copy(values))
    for row_i, row in enumerate(values):
        nans = ~np.isfinite(row)
        if times is None:
            times = np.arange(values.shape[1])

        if max_interp:
            nan_beginnings = np.where(np.diff(nans.astype(int)) == 1)[0] + 1
            nan_endings = np.where(np.diff(nans.astype(int)) == -1)[0] + 1
            if nans[0]:
                nan_beginnings = np.concatenate(([0], nan_beginnings))
            if nans[-1]:
                nan_endings = np.concatenate((nan_endings, [len(nans) - 1]))

            nan_lengths = nan_endings - nan_beginnings
            for episode_i, nan_length in enumerate(nan_lengths):
                if nan_length > max_interp:
                    start_i = nan_beginnings[episode_i]
                    nans[start_i : start_i + nan_length] = False

        if np.any(nans):
            values[row_i, nans] = np.interp(
                times[nans], times[~nans], row[~nans]
            )
    if pad_periodic:
        values = values[:, half:-half]
    return np.squeeze(values)


def sumup(
    values,
    width=24,
    times_=None,
    drop_extra=True,
    mean=False,
    middle_time=True,
    sum_to_nan=False,
    acceptable_nans=6,
    max_interp=3,
):
    """Sum up width number of values along the rows. If there are surplus
    entries, they are dropped as if they were hot (Snoop Dog et al).

    Examples
    --------
    >>> import numpy as np
    >>> a = np.arange(10.).reshape((2, 5))
    >>> a
    array([[ 0.,  1.,  2.,  3.,  4.],
           [ 5.,  6.,  7.,  8.,  9.]])
    >>> sumup(a, 2)
    array([[  1.,   5.],
           [ 11.,  15.]])
    >>> sumup(a, 2, drop_extra=False)
    array([[  1.,   5.,   8.],
           [ 11.,  15.,  18.]])
    >>> sumup(a.ravel(), 5)
    array([ 10.,  35.])
    >>> a[0, 0] = np.nan
    >>> sumup(a, 2, mean=True)
    array([[ 1. ,  2.5],
           [ 5.5,  7.5]])
    """
    width = int(width)
    if max_interp > 0 and not sum_to_nan:
        values = interp_nonfin(values, max_interp=max_interp)
    if len(values.shape) == 1:
        values = values[np.newaxis, :]
    # we hack the values into a (x, width) shape, sum along the rows and
    # reshape it back
    orig_rows, orig_columns = values.shape
    surplus_columns = orig_columns % width
    if drop_extra and surplus_columns:
        values = values[:, :-surplus_columns]
        orig_columns -= surplus_columns
    elif (not drop_extra) and surplus_columns:
        last_values_mean = values[:, np.newaxis, -surplus_columns:].mean(
            axis=2
        )
        last_values_mean = np.array(last_values_mean, dtype=values.dtype)
        values = np.concatenate((values, last_values_mean), axis=1)
    values = values.reshape((values.size // width, width))
    summed_values = np.nansum(values, axis=1)

    if np.sum(np.isnan(values)) > 0:
        nan_counts = np.sum(np.isnan(values), axis=1)
        if sum_to_nan:
            summed_values[nan_counts > 0] = np.nan
        else:
            nan_ii = (nan_counts > 0) & (nan_counts <= acceptable_nans)
            summed_values[nan_ii] *= width / (
                float(width) - nan_counts[nan_ii]
            )
            summed_values[nan_counts > acceptable_nans] = np.nan
    if mean:
        summed_values = summed_values.astype(float)
        summed_values /= width - np.sum(np.isnan(values), axis=1)

    new_columns = int(np.ceil(float(orig_columns) / width))
    summed_values = summed_values.reshape((orig_rows, new_columns))

    if times_ is not None:
        if middle_time:
            # use the time in the middle between the data points
            time_shift = round(width / 2.0)
        else:
            time_shift = None
        times_ = times_[time_shift::width][: summed_values.shape[1]]
        return np.squeeze(summed_values), times_
    else:
        return np.squeeze(summed_values)


def gaps(data):
    """Return indices referring to start and end points of gaps (marked by nans
    in the given array 'data'

    Parameters
    ----------
    data : 1dim ndarray, dtype float or bool

    >>> import numpy as np
    >>> a = np.arange(20.)
    >>> a[-1] = np.nan
    >>> gaps(a)
    array([[19, 19]])
    >>> a = np.arange(20.)
    >>> a[[0, 3, 4, 5, 12, 13, -1]] = np.nan
    >>> gaps(a)
    array([[ 0,  0],
           [ 3,  5],
           [12, 13],
           [19, 19]])
    >>> a = np.arange(20.)
    >>> a[[0, 1, 3, 4, 5, 12, -2, -1]] = np.nan
    >>> gaps(a)
    array([[ 0,  1],
           [ 3,  5],
           [12, 12],
           [18, 19]])
    """
    if data.dtype == bool:
        mask = data
    else:
        mask = np.isnan(data)
    if np.all(~mask):
        return []
    if np.all(mask):
        return [[0, len(mask) - 1]]
    diff = np.diff(mask.astype(int))
    begin_ii = (np.where(diff == 1)[0] + 1).tolist()
    end_ii = np.where(diff == -1)[0].tolist()
    if mask[0]:
        if len(begin_ii) == 0:
            begin_ii = [0]
        if not end_ii[0] < begin_ii[0]:
            end_ii = [0] + end_ii
        begin_ii = [0] + begin_ii
    if (begin_ii[-1] and not end_ii) or begin_ii[-1] > end_ii[-1]:
        end_ii += [len(mask) - 1]

    return np.array([begin_ii, end_ii]).T


def list_transpose(list_):
    """Transposes a "2-dim" nested list.

    Examples
    --------
    >>> list_transpose([[1, 2, 3], [4, 5, 6]])
    [[1, 4], [2, 5], [3, 6]]
    """
    return list(map(list, list(zip(*list_))))


def csv2list(
    filename,
    startfrom=None,
    delimiter=None,
    column_ids=None,
    conversions=None,
    comment="#",
):
    """Returns a list of each column of a csv-file."""
    if conversions is None:
        # no conversion corresponds to a string-conversion.  itertools.repeat
        # gives us those for as many columns that might be there.
        conversions = itertools.repeat(str)
    if (len(np.atleast_1d(column_ids)) > 1) and (
        len(np.atleast_1d(conversions)) == 1
    ):
        # as convenience, this expands the conversions to the length of the
        # given column_ids
        column_conversions = conversions
        conversions = itertools.repeat(conversions)

    with open(filename) as csv_file:
        all_data = [
            [
                conversion(value.strip())
                for value, conversion in zip(
                    line.split(delimiter), conversions
                )
            ]
            for line in itertools.islice(csv_file, startfrom)
            if not line.lstrip().startswith(comment)
        ]

    # "transpose" rows to columns
    columns = list_transpose(all_data)

    if column_ids is None:
        return columns
    elif len(column_ids) == 1:
        # do not return a nested list if there is only one column
        return [
            column_conversions[0](value) for value in columns[column_ids[0]]
        ]
    else:
        return [
            [conversion(value) for value in columns[ii]]
            for ii, conversion in zip(column_ids, column_conversions)
        ]


def csv2dict(filename, *args, **kwds):
    """Returns a dictionary containing the columns of a csv-file.  The keys
    are taken from the first row of the file."""
    aslist = csv2list(filename, *args, **kwds)
    return {column[0].strip(): column[1:] for column in aslist}


# Plotting


class LegendSubtitleHandler(object):
    def legend_artist(self, legend, orig_handle, fontsize, handlebox):
        # this dedents the label
        handlebox.set_width(0)
        # a dummy with zero width and no visible edge
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        patch = mpatches.Rectangle(
            [x0, y0],
            0,
            handlebox.height,
            edgecolor=(0, 0, 0, 0),
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


legend_subtitle = LegendSubtitleHandler()


def square_subplots(n_variables, *args, **kwds):
    """Similar to plt.subplots, but shares x-axes column-wise and y-axes
    row - wise. Main diagonal subplots only share x-axes."""
    fig = plt.figure(*args, **kwds)
    axes = np.empty((n_variables, n_variables), dtype=object)
    for ii in range(n_variables):
        for jj in range(n_variables):
            if ii == 0:
                sharex_ax = None
            else:
                sharex_ax = axes[0, jj]

            if (jj == 0) or (ii == jj):
                sharey_ax = None
            elif ii == 0:
                sharey_ax = axes[ii, 1]
            else:
                sharey_ax = axes[ii, 0]

            axes[ii, jj] = fig.add_subplot(
                n_variables,
                n_variables,
                ii * n_variables + jj + 1,
                sharex=sharex_ax,
                sharey=sharey_ax,
            )
    return fig, axes


def splom(
    data,
    variable_names=None,
    f_kwds=None,
    h_kwds=None,
    s_kwds=None,
    opacity=0.1,
    highlight_mask=None,
    ticklabels=True,
    figsize=None,
    hists=True,
    facecolor=(0, 0.5, 0.5),
    edgecolor=(1, 1, 1),
    f_opacity=None,
    e_opacity=None,
    highlight_color="red",
):
    """Scatter-plot matrix with interactive capabilities.

    Parameters
    ----------
    data : (K,T) ndarray
        K variables, T timesteps
    variable_names : sequence of strings, optional
        Used to label the subplots.
    f_kwds : dictionary, optional
        Keyword arguments that are passed to plt.subplots.
    h_kwds : dictionary, optional
        Keyword arguments that are passed to the histogramm calls.
    s_kwds : dictionary, optional
        Keyword arguments that are passed to the scatter calls.
    opacity : float, optional
        Opacity used for edgecolor parameter of scatter call.
    highlight_mask : (T,) boolean ndarray, optional
        Mask of timesteps that will be highlighted in the scatter plots.
    ticklabels : boolean, optional
        Set to False to supress displaying x- and yticklabels.
    figsize : None or tuple of width, height, optional
        Size of the figure.
    hists : boolean, optional
        Plot histograms on the main diagonal.
    """
    cc = mpl.colors.ColorConverter()

    def switch_fc(artist, ind):
        fc = artist._facecolors
        if len(fc) == 1:
            fc = np.array(len(artist._offsets) * fc.tolist())
        fc[ind] = 1 - fc[ind]
        return fc

    def brush(event):
        """Highlight points in all plots."""
        # cache the collections. we will hopefully not get any more subplots
        if not hasattr(brush, "collections"):
            brush.collections = [
                sub.collections[0]
                for sub in fig.get_children()
                if hasattr(sub, "collections") and len(sub.collections) == 1
            ]
        # handle the click on a histogram bin
        if type(event.artist) is mpl.patches.Rectangle:
            event.artist._facecolor = tuple(
                1 - color_comp for color_comp in event.artist._facecolor
            )
            # find the indices of the points within the bin clicked on
            lower = event.artist._x
            upper = lower + event.artist._width
            values = data[event.artist.axes.var_i].ravel()
            ind = np.where((values > lower) & (values <= upper))
        else:
            # handle the click on a single scatter point
            ind = event.ind
        for col in brush.collections:
            col._facecolors = switch_fc(col, ind)
        fig.canvas.draw()

    f_kwds = {} if f_kwds is None else f_kwds
    h_kwds = {} if h_kwds is None else h_kwds
    s_kwds = {} if s_kwds is None else s_kwds
    f_opacity = opacity if f_opacity is None else f_opacity
    e_opacity = opacity if e_opacity is None else e_opacity
    data = np.asarray(data)
    n_variables = data.shape[0]
    fig, axes = square_subplots(n_variables, figsize=figsize, **f_kwds)

    for ii in range(n_variables):
        for jj in range(n_variables):
            if ii == jj:
                if hists:
                    axes[ii, jj].hist(
                        np.where(np.isnan(data[ii]), 0, data[ii]),
                        min(20, int(len(data[ii]) ** 0.5)),
                        picker=5,
                        density=True,
                        # want to achieve red when inverting
                        facecolor=cc.to_rgba(facecolor, alpha=0),
                        **h_kwds,
                    )
                    # store the ii-index as an attribute to identify the
                    # variable later in the brush function
                    axes[ii, jj].var_i = ii
                else:
                    fig.delaxes(axes[ii, jj])
            else:
                if highlight_mask is None:
                    facecolors = cc.to_rgba(facecolor, alpha=f_opacity)
                else:
                    facecolors = np.empty((data.shape[1], 4))
                    facecolors[highlight_mask] = cc.to_rgba(
                        highlight_color, f_opacity
                    )
                    # want to achieve red when inverting
                    facecolors[~highlight_mask] = cc.to_rgba(
                        facecolor, f_opacity
                    )
                axes[ii, jj].scatter(
                    data[jj],
                    data[ii],
                    picker=5,
                    facecolors=facecolors,
                    edgecolors=cc.to_rgba(edgecolor, e_opacity),
                    **s_kwds,
                )
            # show ticklabels only on the margins
            #                if (jj != 0) or (ii == jj):
            #                    axes[ii, jj].set_yticklabels("")
            #                if ii != n_variables - 1:
            #                    axes[ii, jj].set_xticklabels("")
            if variable_names and jj == 0:
                axes[ii, jj].set_ylabel(variable_names[ii])
            if variable_names and ii == n_variables - 1:
                axes[ii, jj].set_xlabel(variable_names[jj])
            if not ticklabels:
                axes[ii, jj].set_yticks([])
                axes[ii, jj].set_xticks([])

    fig.canvas.mpl_connect("pick_event", brush)
    return fig


def kde_gauss(
    dataset,
    evaluation_points=None,
    kernel_width=None,
    maxopt=500,
    return_width=False,
    verbose=False,
):
    """

    Parameters
    ----------
    dataset : (T,) ndarray
        input data as array of length T
    evaluation points : (N,) ndarray, optional if return_width=True
        N points as array (e.g. xx=np.linspace(-3,3,100)
    kernel_width : float, optional
        kernel_width (eg 0.3) if set, no MLM-routine to infer optimal kernel
        width
    maxopt : int, optional
        size of sample to optimize kernel_width, affects runtime strongly.
        depends on memorysize. max 1000 with 2GB mem
    return_width : boolean, optional
        Return the optimized kernel width and nothing else.
    verbose : boolean, optional
        Print information (also from the optimizer).

    Examples
    --------
        dataset = np.array(np.random.normal(size=1e3))
        xx = np.linspace(-3,3, 1e3)
        plt.plot(xx,kde_gauss(dataset,xx))
    """

    def residual_matrix(x_vec, y_vec):
        """Returns a matrix with residuals (x:horizontal,y:vertical)"""
        x_vec, y_vec = list(map(np.asarray, (x_vec, y_vec)))
        return x_vec[None, :] - y_vec[:, None]

    def neglog_likelihood(kernel_width, dataset, verbose=False):
        """optimizing kernel width with MLM leave one out"""

        optMatrix = residual_matrix(dataset, dataset)
        if NE:
            pi = np.pi
            ne_str = (
                "1.0 / (sqrt(2 * pi) * kernel_width) / "
                "exp(optMatrix ** 2 / (2 * kernel_width ** 2))"
            )
            optMatrix = ne.evaluate(ne_str)
        else:
            preTerm = 1.0 / (np.sqrt(2 * np.pi) * kernel_width)
            optMatrix = preTerm / np.exp(optMatrix**2 / (2 * kernel_width**2))
        nDataset = np.shape(dataset)[0]
        # sets diagonal to 0, i.e. leave-one-out method
        optMatrix.ravel()[:: nDataset + 1] = 0
        densities = np.sum(optMatrix, axis=1) / float(nDataset - 1)
        # LN if <>0 for MLM
        d_sum = 0
        err = 0
        for d in densities:
            if d > 0:
                d_sum -= np.log(d)
            else:
                if err == 0:
                    if verbose:
                        print("LN(0) case do attend")
                    err = 1
                d_sum += 100  # not nice
        return d_sum

    dataset = np.asarray(dataset)
    # problem bei optimierung: d fluktuiert und haengt von maxopt ab
    # je hoeher maxopt desto kleiner d!
    dataset = np.sort(dataset)

    # optimizing kernel width if d=None
    if kernel_width is None:
        data_width = dataset.max() - dataset.min()
        d_0 = data_width / 10
        if d_0 < 0.0001:
            d_0 = 0.0001
        if len(dataset) > 1000:
            fluct = True  # while values fluctuate, repeat iteration
            d_n = []  # list of d's
            d_act, d_old = d_0, 0
            n_min = 8 + np.sqrt(len(dataset) / maxopt)  # min nr of iterations
            n_act = 0

            while fluct or n_act <= n_min:
                dataset_sample = random.sample(tuple(dataset), maxopt)
                d_n.append(
                    optimize.fmin(
                        neglog_likelihood,
                        d_act,
                        args=(dataset_sample, verbose),
                        disp=verbose,
                    )[0]
                )
                d_act = sum(d_n) / float(len(d_n))
                # stop if fluct < 1%
                if abs(d_act - d_old) / float(d_act) < 0.01:
                    fluct = False
                if verbose:
                    print(d_act, d_old, d_n[-1])
                d_old = d_act
                n_act += 1
            kernel_width = d_act

        else:
            dataset_sample = dataset
            kernel_width = optimize.fmin(
                neglog_likelihood,
                d_0,
                args=(dataset_sample, verbose),
                disp=verbose,
            )
        if verbose:
            print("Kernelwidth = %f" % kernel_width)

    if return_width:
        return kernel_width

    evaluation_points = np.asarray(evaluation_points)

    if len(dataset) < len(evaluation_points):
        print(
            "Caution: you get more ev. points than input data\
        be aware of pseudo exactness"
        )

    # save kernel width, so it can be retrieved if anyone is interested
    kde_gauss.kernel_width = kernel_width

    # creating Matrix with residuals
    #    kdeMatrix = residual_matrix(dataset,evaluation_points)
    #    sparse_kde_mask = kdeMatrix < .001
    #    from scipy.sparse import lil_matrix
    #    sparse_kde = lil_matrix(kdeMatrix.shape)
    #    sparse_kde[sparse_kde_mask] = kdeMatrix[sparse_kde_mask]
    #     using Gaussian kernel
    #    import numexpr as ne
    #    preTerm = ne.evaluate("1.0 / ((2 * math.pi)**.5 * kernel_width)")
    #    kdeMatrix = preTerm / np.exp(kdeMatrix ** 2 / (2 * kernel_width ** 2))

    if len(dataset) * len(evaluation_points) > 1e7:
        parts = int(len(dataset) * len(evaluation_points) / 1e7) + 1
        brIncr = int(len(evaluation_points) / parts)
        densities = np.array([])
        for i in range(parts + 1):
            if verbose:
                print("part %i of %i" % (i, parts))
            kdeMatrix = residual_matrix(
                dataset, evaluation_points[i * brIncr : (i + 1) * brIncr]
            )
            preTerm = 1.0 / (np.sqrt(2 * np.pi) * kernel_width)
            kdeMatrix = (
                preTerm / np.exp(kdeMatrix**2 / (2 * kernel_width**2)),
            )
            tmp_densities = np.sum(kdeMatrix, axis=1) / float(len(dataset))
            densities = np.hstack((densities, tmp_densities))
    else:
        kdeMatrix = residual_matrix(dataset, evaluation_points)
        preTerm = 1.0 / (np.sqrt(2 * np.pi) * kernel_width)
        with np.errstate(all="ignore"):
            kdeMatrix = preTerm / np.exp(kdeMatrix**2 / (2 * kernel_width**2))
        # suming lines
        densities = np.sum(kdeMatrix, axis=1) / float(len(dataset))
    return densities


def hist(
    values,
    n_bins,
    dist=None,
    pdf=None,
    kde=False,
    fig=None,
    ax=None,
    discrete=False,
    figsize=None,
    legend=True,
    *args,
    **kwds,
):
    """Plots a histogram and therotical or empirical densities."""
    try:
        if np.any(~np.isfinite(values)):
            warnings.warn("Non-finite values in values.")
    except TypeError:
        pass
    figsize = plt.rcParams["figure.figsize"] if figsize is None else figsize
    if ax is None:
        fig = plt.figure(figsize=figsize) if fig is None else fig
        axes = ax1 = fig.add_subplot(111)
    else:
        axes = ax1 = ax

    # the histogram of the data
    if discrete:
        values_2d = np.atleast_2d(values)
        bin_offset = -0.5 * values_2d.shape[0]
        for i, values in enumerate(values_2d):
            values = np.array(values)
            bins = np.arange(values.min(), values.max() + 1, dtype=int)
            bins = bins + bin_offset + i
            freqs = np.bincount(values.astype(int))
            freqs = freqs[freqs >= bins.min()]
            freqs = freqs.astype(float) / values.size
            ax1.vlines(bins, 0, freqs, linewidth=3)
            ax1.set_xlim(bins[0] - 1, bins[-1] + 1)
    else:
        bins = ax1.hist(
            values,
            n_bins,
            density=True,
            facecolor="grey",
            alpha=0.75,
            *args,
            **kwds,
        )[1]

    ax1.set_ylabel("relative frequency")

    if not (isinstance(values, list) or values.ndim == 2):
        values_2d = (values,)
    else:
        values_2d = values

    if discrete:
        eva_points = bins
    else:
        eva_points = np.linspace(bins[0], bins[-1], 4 * n_bins)
    if kde:
        for val_i, values in enumerate(values_2d):
            density = kde_gauss(values, eva_points)
            ax1.plot(eva_points, density, label=("kde%d" % val_i))
    if dist:
        try:
            dist[0]
            dists = dist
        except TypeError:
            dists = (dist,)

        # the quantile part
        ax2 = ax1.twinx()
        axes = [ax1, ax2]
        for values in values_2d:
            # empirical cdf
            values_sort = np.sort(values)
            ranks_emp = (0.5 + np.arange(len(values))) / len(values)
            ax2.plot(values_sort, ranks_emp)
            pdf = []
            for dist in dists:
                if hasattr(dist, "fit"):
                    fitted_dist = dist(*dist.fit(values))
                else:
                    fitted_dist = dist
                pdf += [fitted_dist.pdf]
                # theoretical cdf
                ranks_theory = fitted_dist.cdf(eva_points)
                p_val = stats.kstest(values, fitted_dist.cdf, mode="asymp")[1]
                ax2.plot(
                    eva_points,
                    ranks_theory,
                    "--",
                    label=("%s p-value: %.1f%%" % (dist.name, p_val * 100)),
                )
                ax2.set_ylabel(r"cumulative frequency")
                ax2.set_ylim(0, 1)
                ax2.grid()

        if len(dists) == 1:
            if hasattr(dist, "parameter_names"):
                plt.title(
                    " ".join(
                        "%s:%.3f" % (par_name, par)
                        for par_name, par in zip(
                            dist.parameter_names, fitted_dist.params
                        )
                    )
                )
            elif hasattr(fitted_dist, "args"):
                plt.title(" ".join("%.3f" % par for par in fitted_dist.args))
        elif len(dists) > 1 and legend:
            plt.legend(loc="best")
    if pdf:
        try:
            pdf[0]
            pdfs = pdf
        except TypeError:
            pdfs = (pdf,)
        for pdf in pdfs:
            density_th = pdf(eva_points)
            if discrete:
                density_th *= len(values)
            ax1.plot(
                eva_points,
                density_th,
                "--o" if discrete else "--",
                linewidth=1,
                label="pdf",
            )

    if fig is not None:
        return fig, axes
    else:
        axes


def scale_yticks(event):
    """Automagically make room for yticklabels.
    Use it like this:
        fig = gcf()
        fig.canvas.mpl_connect('draw_event', scale_yticks)

    Stolen from the matplotlib-howto.  Slightly changed, so it is possible to
    separate the function from the calling code.
    http://matplotlib.sourceforge.net/faq/howto_faq.html\
    #automatically-make-room-for-tick-labels
    """
    labels = plt.gca().get_yticklabels()
    fig = plt.gcf()
    bboxes = []
    for label in labels:
        bbox = label.get_window_extent()
        # the figure transform goes from relative coords->pixels and we
        # want the inverse of that
        # bboxi = bbox.inverse_transformed(fig.transFigure)
        bboxi = bbox.transformed(fig.transFigure.inverted())
        bboxes.append(bboxi)

    # this is the bbox that bounds all the bboxes, again in relative
    # figure coords
    bbox = mpl.transforms.Bbox.union(bboxes)
    if fig.subplotpars.left < bbox.width:
        # we need to move it over
        fig.subplots_adjust(left=1.1 * bbox.width)  # pad a little
        fig.canvas.draw()
    return False


def yscale_subplots(fig=None, per_type=False, regrid=False):
    """Sets a common y-scale to all subplots.  If per_type is set to True,
    y-scales are distinguished by the type of the subplots."""
    if fig is None:
        fig = plt.gcf()

    if per_type:
        key_func = type
    else:

        def key_func(x):
            return "the one to rule them all"

    # see http://matplotlib.sourceforge.net/faq/howto_faq.html#\
    # find-all-objects-in-figure-of-a-certain-type
    def ylim_getable(sub):
        return hasattr(sub, "get_ylim")

    def ylim_setable(sub):
        return hasattr(sub, "set_ylim")

    # find the y-limits of each subplot
    ymins, ymaxs = {}, {}
    for subplot in fig.findobj(ylim_getable):
        ymin, ymax = subplot.get_ylim()
        sub_type = key_func(subplot)
        if sub_type not in ymins:
            ymins[sub_type], ymaxs[sub_type] = [], []
        ymins[sub_type].append(ymin)
        ymaxs[sub_type].append(ymax)

    # find the extremes for each type of subplot
    ymin, ymax = {}, {}
    for sub_type in list(ymins.keys()):
        ymin[sub_type] = min(ymins[sub_type])
        ymax[sub_type] = max(ymaxs[sub_type])

    # set ylims (and reset grids)
    for subplot in fig.findobj(ylim_setable):
        sub_type = key_func(subplot)
        try:
            subplot.set_ylim(ymin=ymin[sub_type], ymax=ymax[sub_type])
        except TypeError:
            pass
        if regrid:
            subplot.set_rgrids(np.linspace(1e-6, ymax[sub_type], 10))


if __name__ == "__main__":
    import doctest

    doctest.testmod()

    fig, ax = plt.subplots()
    (line,) = ax.plot(np.arange(10))
    ax.legend(
        [line, None],
        ("line", "find me"),
        handler_map={None, LegendSubtitleHandler()},
    )
    plt.show()



import warnings
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import varwg
from varwg import helpers as my


def trans_prob(rain, thresh=0.0001):
    """Estimates transition probabilities.

    Parameters
    ---------
    rain : 1d array
    thresh : float, optional
        threshold a value must exceed to be interpreted as rain [m]

    Returns
    -------
    trans : (2, 2) ndarray
        Transition probabilities.
        [[P(no rain -> no rain), P(no rain -> rain],
         [P(rain -> no rain),    P(rain -> rain)]]
    """
    rain_mask = rain >= thresh
    n = float(len(rain))
    diffs = np.diff(rain_mask.astype(int))
    no_change_mask = diffs == 0
    p11 = np.sum(rain_mask[:-1] & no_change_mask)
    p00 = np.sum(~rain_mask[:-1] & no_change_mask)
    p10 = np.sum(diffs == -1)
    p01 = np.sum(diffs == 1)
    return np.array([[p00, p01], [p10, p11]]) / n


def spell_lengths(rain, thresh=0.001):
    """Calculate dry and wet spell lenghts.

    Parameters
    ----------
    rain : 1d array
    thresh : float, optional
        threshold a value must exceed to be interpreted as rain [m]

    Returns
    -------
    dry : 1d int array
    wet : 1d int array
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        wet_mask = rain >= thresh
        # dry_mask = rain <= 0
    dry = np.array(
        [stop_i - start_i + 1 for start_i, stop_i in my.gaps(~wet_mask)]
    )
    wet = np.array(
        [stop_i - start_i + 1 for start_i, stop_i in my.gaps(wet_mask)]
    )
    return dry, wet


def plot_exceedance(
    obs,
    sim=None,
    kind="depth",
    thresh=0.001,
    fig=None,
    axs=None,
    figsize=None,
    lkwds=None,
    draw_scatter=None,
    log="both",
    *pargs,
    **pkwds,
):
    """Plot exceedance probability.

    Parameters
    ---------
    obs : 1-dim ndarray
        observations
    sim : 1 or 2-dim ndarray, optional
        simulated values. If 2-dimensional, median, and range of those values
        will be plotted.
    kind : "depth", "dry", "wet" or "all", optional
        What to plot:
        - depth: rain intensity
        - dry: lengths of dry spells
        - wet: lengths of wet spells
        - all
    thresh : float, optional
        threshold a value must exceed to be interpreted as rain [m].
        only needed if kind != "depth".
    """
    if draw_scatter is None:
        draw_scatter = not bool(fig)
    if kind == "all":
        if fig is None:
            fig, axs = plt.subplots(
                ncols=3, sharey=True, figsize=figsize, constrained_layout=True
            )
        kinds = "depth", "dry", "wet"
    elif isinstance(kind, str):
        kinds = (kind,)
    elif fig is None or axs is None:
        fig, axs = plt.subplots(
            figsize=figsize, sharey=True, constrained_layout=True
        )
        axs = (axs,)
        kinds = (kind,)
    if lkwds is None:
        lkwds = {}

    def rel_ranks(n):
        return (np.arange(float(n)) + 0.5) / n

    def plot_sim_single(ax, sim):
        sim_conv = conv(sim)
        ranks = rel_ranks(len(sim_conv))
        sim_sorted = np.sort(sim_conv)[::-1]
        match log:
            case "x":
                plot_func = ax.semilogx
            case "y":
                plot_func = ax.semilogy
            case "both":
                plot_func = ax.loglog
            case False | "none":
                ax.plot
            case _:
                raise RuntimeError(
                    "log parameter must be one of: 'x', 'y', 'both', 'none', False"
                )
        plot_func(sim_sorted, ranks, *pargs, **pkwds)

    edgecolor = lkwds.pop("edgecolor", "black")
    for ax_i, kind in enumerate(kinds):
        ax = axs[ax_i]

        if kind == "depth":

            def conv(x):
                valid_x = x[np.isfinite(x)]
                return valid_x[valid_x >= thresh]

            ax.set_xlabel("depth [mm]")
        elif kind == "dry":

            def conv(x):
                return spell_lengths(x, thresh)[0]

            ax.set_xlabel("dry spell length")
        elif kind == "wet":

            def conv(x):
                return spell_lengths(x, thresh)[1]

            ax.set_xlabel("wet spell length")

        if np.ndim(sim) == 1:
            plot_sim_single(ax, sim)

        if np.ndim(sim) == 2:
            sim = np.asarray(sim)
            for sim_ in sim:
                plot_sim_single(ax, sim_)

        if draw_scatter:
            obs_conv = conv(obs)
            ranks = rel_ranks(len(obs_conv))
            obs_sorted = np.sort(obs_conv)[::-1]
            ax.scatter(
                obs_sorted,
                ranks,
                marker="o",
                edgecolor=edgecolor,
                # edgecolor=(1, 0, 0, .5),
                facecolor=(0, 0, 0, 0),
                **lkwds,
            )
            ax.set_xscale("log")
            ax.set_yscale("log")
            ax.set_ylim(max(5e-4, ranks[0]), ranks[-1])
            # ax.set_ylim(ranks[0], ranks[-1])
            ax.grid(True)
    axs[0].set_ylabel("1 - F(X)")
    # fig.tight_layout(rect=(0, .1, 1, 1))
    return fig, axs


def richardson_model_occ(T, trans_pp):
    """Simulate rain occurrence with Richardson's (1981) model (2 state, first
    order Markov Chain).

    Parameters
    ----------
    T : int
        Length of simulated time series
    trans_pp : (2, 2) ndarray
        Transition probabilities (see trans_prob).
        [[P(no rain -> no rain), P(no rain -> rain],
         [P(rain -> no rain),    P(rain -> rain)]]

    Returns
    -------
    occs : (T,) bool ndarray
        Rain occurrence.

    References
    ----------
    Richardson, C. W. (1981). Stochastic simulation of daily precipitation,
    temperature, and solar radiation. Water Resources Research, 17(1),
    182-190. doi:10.1029/WR017i001p00182

    See also
    --------
    trans_prob : Estimates transition probabilities given a time series.
    """
    pp = trans_pp
    T = int(round(T))
    occs = np.full(T, False, dtype=bool)
    rr = varwg.rng.random(T)
    # decide on the first occurrence with unconditional probability
    occs[0] = rr[0] < pp[0].sum()
    for t in range(1, T):
        occs[t] = rr[t] < (
            pp[int(occs[t - 1]), 1] / pp[:, int(occs[t - 1])].sum()
        )
    return occs


def richardson_model(T, rain, thresh=0.0001):
    if T is None:
        T = len(rain)
    trans_pp = trans_prob(rain)
    occs = richardson_model_occ(T, trans_pp)
    rain_sim = np.zeros(int(round(T)))
    rain_sim[occs] = varwg.rng.choice(rain[rain > thresh], np.sum(occs))
    return rain_sim


def get_hyd_years_mask(times):
    return (times.month == 10) & (times.day == 1) & (times.hour == 0)


def hyd_year_sums(
    data_xar, fun_name="sum", full_years=True, hyd_years_mask=None
):
    time = data_xar.time
    if hyd_years_mask is None:
        hyd_years_mask = get_hyd_years_mask(time.dt)
    hyd_year = hyd_years_mask.cumsum() - 1
    hyd_year += hyd_year.time[0].dt.year
    if fun_name == "sum":
        summed = data_xar.groupby(hyd_year).sum("time").rename(group="hydyear")
    elif fun_name == "mean":
        summed = (
            data_xar.groupby(hyd_year).mean("time").rename(group="hydyear")
        )
    time_first = time[0]
    month = int(time_first.dt.month)
    day = int(time_first.dt.day)
    hyd_year_coord = pd.date_range(
        f"{summed.hydyear.values[0]}-{month}-{day}",
        f"{summed.hydyear.values[-1] + 1}-{month}-{day}",
        freq="Y",
    )
    summed = summed.assign_coords(dict(hydyear=hyd_year_coord))
    if full_years:
        if not hyd_years_mask[0]:
            summed = summed.isel(hydyear=slice(1, None))
        time_last = time[-1]
        if not ((time_last.dt.month == 9) & (time_last.dt.day == 30)):
            summed = summed.isel(hydyear=slice(None, -1))
    return summed


if __name__ == "__main__":
    import varwg

    varwg.conf = varwg.config_konstanz
    met_vg = varwg.VG(("R", "theta"))
    rain = met_vg.data_raw[0] / 24
    plot_exceedance(rain, kind="all")
    plt.show()



import os
import numpy as np
import numpy.testing as npt
import pytest
import varwg
from varwg.time_series_analysis import models, tests


# the following parameters and data are taken from p.707 in order to test
# functions regarding the VAR least-squares estimator
B_test = np.array(
    [
        [-0.017, -0.320, 0.146, 0.961, -0.161, 0.115, 0.934],
        [0.016, 0.044, -0.153, 0.289, 0.050, 0.019, -0.010],
        [0.013, -0.002, 0.225, -0.264, 0.034, 0.355, -0.022],
    ]
)
Bex_test = np.array(
    [
        [-0.017, -0.320, 0.146, 0.961, -0.161, 0.115, 0.934, 0.25],
        [0.016, 0.044, -0.153, 0.289, 0.050, 0.019, -0.010, 0.1],
        [0.013, -0.002, 0.225, -0.264, 0.034, 0.355, -0.022, -0.25],
    ]
)
A_test = np.array(
    [
        [-0.319, 0.147, 0.959, -0.160, 0.115, 0.932],
        [0.044, -0.152, 0.286, 0.050, 0.020, -0.012],
        [-0.002, 0.225, -0.264, 0.034, 0.355, -0.022],
    ]
)
sigma_u_test = 1e-4 * np.array(
    [[21.3, 0.72, 1.23], [0.72, 1.37, 0.61], [1.23, 0.61, 0.89]]
)
VAR_p = 2
VAR_K = 3
VARMA_p, VARMA_q = 2, 1
e1_filepath = os.path.join(os.path.dirname(tests.__file__), "e1.dat")
data = np.loadtxt(e1_filepath, skiprows=7).T
data = np.log(data[:, 1:]) - np.log(data[:, :-1])
data = data[:, :75]
data_means = data.mean(axis=1).reshape((3, 1))


class Test(npt.TestCase):
    msg = "Values differ from Luetkepohl's. Same p, however."
    # Test is not conclusive: solutions are for ML estimator of VAR process...
    # We get however, the same p!

    @pytest.mark.skip(reason=msg)
    def test_AIC(self):
        """See p.148"""
        order_data = np.copy(data[:, 4:])
        p = 1
        K, T = order_data.shape
        sigma_u = (float(T - K * p - 1) / T) * models.VAR_LS(order_data, p)[1]
        print(np.linalg.det(sigma_u) * 1e11)
        print(models.VAR_LS(order_data, p)[1] * 1e4)
        objectives = np.array(
            [
                (float(T - K * p_ - 1) / T)
                * models.AIC(
                    models.VAR_LS(order_data, p_)[1], p_, order_data.shape[1]
                )
                for p_ in range(5)
            ]
        )
        AICs = -24 - np.array([0.42, 0.5, 0.59, 0.41, 0.36])
        npt.assert_almost_equal(AICs, objectives, decimal=2)

    @pytest.mark.skip(reason=msg)
    def test_HQ(self):
        order_data = np.copy(data[:, 4:])
        objectives = np.array(
            [
                models.HQ(
                    models.VAR_LS(order_data, p)[1], p, order_data.shape[1]
                )
                for p in range(5)
            ]
        )
        HQs = -24 - np.array([0.42, 0.38, 0.37, 0.07, -0.1])
        npt.assert_almost_equal(HQs, objectives, decimal=2)

    @pytest.mark.skip(reason=msg)
    def test_SC(self):
        order_data = np.copy(data[:, 4:])
        objectives = np.array(
            [
                models.SC(
                    models.VAR_LS(order_data, p)[1], p, order_data.shape[1]
                )
                for p in range(5)
            ]
        )
        SCs = np.array([-24.42, -24.21, -24.02, -23.55, -23.21])
        npt.assert_almost_equal(SCs, objectives, decimal=2)

    @pytest.mark.skip(reason=msg)
    def test_FPE(self):
        order_data = np.copy(data[:, 4:])
        K, T = order_data.shape
        objectives = np.array(
            [
                (float(T - K * p - 1) / T)
                * models.FPE(models.VAR_LS(order_data, p)[1], p, T)
                for p in range(5)
            ]
        )
        objectives *= 1e11
        FPEs = np.array([2.691, 2.5, 2.272, 2.748, 2.91])
        npt.assert_almost_equal(FPEs, objectives, decimal=2)

    def test_B2A(self):
        B = np.array([[0, 1, 2, 5, 6], [0, 3, 4, 7, 8]])
        A_exp = np.array(
            [[1, 2, 5, 6], [3, 4, 7, 8], [1, 0, 0, 0], [0, 1, 0, 0]]
        )
        A_act = models.B2A(B)
        npt.assert_almost_equal(A_act, A_exp)

    # def test_VAR_cov(self):
    #     # VAR(1)
    #     B = np.array([[0, 0.5, 0, 0], [0, 0.1, 0.1, 0.3], [0, 0, 0.2, 0.3]])
    #     sigma_u = np.array([[2.25, 0, 0], [0, 1, 0.5], [0, 0.5, 0.74]])
    #     cov_exp = np.array(
    #         [[3, 0.161, 0.019], [0.161, 1.172, 0.674], [0.019, 0.674, 0.954]]
    #     )
    #     cov_act = models.VAR_cov(B, sigma_u)
    #     npt.assert_almost_equal(cov_act, cov_exp, decimal=3)
    #     # VAR(2)
    #     B = np.array([[0, 0.5, 0.1, 0, 0], [0, 0.4, 0.5, 0.25, 0]])
    #     sigma_u = np.array([[0.09, 0], [0, 0.04]])
    #     cov_exp = np.array([[0.131, 0.066], [0.066, 0.181]])
    #     cov_act = models.VAR_cov(B, sigma_u)
    #     npt.assert_allclose(cov_act, cov_exp, atol=1e-3)

    def test__scale_additive(self):
        self.assertRaises(
            ValueError,
            models._scale_additive,
            [1, 0],
            [[0.5, 0], [0, 1], [0, 0]],
        )

        scaled = models._scale_additive([1, 0], [[0.5, 0], [0, 1]])
        self.assertAlmostEqual(tuple(scaled), (0.5, 0))

        scaled = models._scale_additive(
            [1, 0], [[0.5, 0, 0.5, 0], [0, 1, 0, 1]], p=1
        )
        self.assertAlmostEqual(tuple(scaled), (0.5, 0))

        scaled = models._scale_additive([[1, 0.5], [0, 0]], [[0.5, 0], [0, 1]])
        npt.assert_almost_equal(scaled, np.array([[0.5, 0.25], [0, 0]]))
        self.assertEqual(scaled.shape, (2, 2))

    def test_VAR_LS(self):
        """Checking the example mentioned in the VAR_LS-docstring (p. 707f)."""
        # values from 1960-1978 with presample data
        B, sigma_u = models.VAR_LS(data, VAR_p, biased=False)
        # sigma_u = np.cov(models.VAR_LS_residuals(data, B, VAR_p), ddof=1)
        npt.assert_almost_equal(B, B_test, decimal=3)
        npt.assert_almost_equal(sigma_u, sigma_u_test, decimal=6)

    # def test_VAR_YW(self):
    #     """Checking the example mentioned in the VAR_LS-docstring (p. 707f)."""
    #     # values from 1960-1978 with presample data
    #     A, sigma_u = models.VAR_YW(data, VAR_p)
    #     print "\n", A
    #     sigma_u = np.cov(models.VAR_LS_residuals(data, B, VAR_p), ddof=1)
    #     assert_almost_equal(A_test, A, decimal=2)

    def test_VAR_LS_sim(self):
        """Does VAR_LS_sim reproduce the correlation matrix?"""
        T = 500
        sim = models.VAR_LS_sim(B_test, sigma_u_test, T)
        npt.assert_almost_equal(np.corrcoef(data), np.corrcoef(sim), decimal=1)

    def test_VAR_LS_sim_m(self):
        """Does VAR_LS_sim change means as requested."""
        T = 500
        varwg.reseed(0)
        sim = models.VAR_LS_sim(B_test, sigma_u_test, T)
        varwg.reseed(0)
        m = np.array([2, 1, 0.5])
        mt = np.empty((len(m), T))
        mt.T[:] = m
        sim_m = models.VAR_LS_sim(B_test, sigma_u_test, T, m=mt)
        m_actual = sim_m.mean(axis=1) - sim.mean(axis=1)
        npt.assert_almost_equal(m_actual, m, decimal=1)

    def test_VAR_LS_sim_fixed_var(self):
        """Does VAR_LS_sim return the fixed values given via fixed_data?"""
        T = 5
        fixed = np.nan * np.empty(VAR_K * T).reshape((VAR_K, T))
        fixed[0] = np.array([0, 0, 1.0, 0, 0])
        sim = models.VAR_LS_sim(B_test, sigma_u_test, T, fixed_data=fixed)
        npt.assert_almost_equal(sim[0], fixed[0])

    def test_VARMA_LS_sim_fixed_var(self):
        """Does VAR_LS_sim return the fixed values given via fixed_data?"""
        T = 5
        fixed = np.nan * np.empty(VAR_K * T).reshape((VAR_K, T))
        fixed[0] = np.array([0, 0, 1.0, 0, 0])
        AM, sigma_u = models.VARMA_LS_prelim(data, VARMA_p, VARMA_q)[:-1]
        sim = models.VARMA_LS_sim(
            AM, VARMA_p, VARMA_q, sigma_u, data_means, T, fixed_data=fixed
        )
        npt.assert_almost_equal(sim[0], fixed[0])

    #  def test_VARMA_LS_sim(self):
    #      """Does VAR_LS_sim reproduce the correlation matrix?"""
    #      T = data.shape[1] * 100
    #      AM, sigma_u = models.VARMA_LS_prelim(data, VARMA_p, VARMA_q)[:-1]
    #      means = data.mean(axis=1)
    #      sim = models.VARMA_LS_sim(AM, VARMA_p, VARMA_q, sigma_u, means, T)
    #      npt.assert_almost_equal(np.corrcoef(data), np.corrcoef(sim), decimal=2)

    # def test_VARMA_LS_sim_VARMA_residuals(self):
    #     """Round-trip test: do we get predifined residuals back from
    #     VARMA_LS_residuals when simulating with VARMA_LS_sim?"""
    #     T = 5
    #     residuals_test = np.asmatrix(
    #                         [varwg.rng.multivariate_normal(VAR_K * [0],
    #                                                        sigma_u_test)
    #                          for t in xrange(T)]).reshape((VAR_K, T))
    #     AM, sigma_u = models.VARMA_LS_prelim(data, VARMA_p, VARMA_q)[:-1]
    #     sim = models.VARMA_LS_sim(AM, VARMA_p, VARMA_q, sigma_u, T,
    #                               residuals_test, n_sim_multiple=1)
    #     residuals = models.VARMA_residuals(sim, AM, VARMA_p, VARMA_q)
    #     npt.assert_almost_equal(residuals, residuals_test, decimal=3)

    # def test_MGARCH_sim_MGARCH_residuals(self):
    #     varwg.reseed(0)
    #     # B = [[-.24e-3, -.00, 0.02, -.18, .16, -.08, .11],
    #     #      [-.42e-3, 0.12, -.13, -.08, .03, -.01, .01]]
    #     # sigma_u = [[1., .8],
    #     #            [.8, 1.]]
    #     T = 1000
    #     sigma_z = .5
    #     a0, a1 = .25, .5
    #     sigma_t = np.ones((2, T))
    #     ut = np.zeros((2, T))
    #     zts = sigma_z * varwg.rng.multivariate_normal([0, 0],
    #                                                   [[1., .8],
    #                                                    [.8, 1.]],
    #                                                   T)
    #     zts = zts.T
    #     for t in xrange(1, T):
    #         sigma_t[:, t] = np.sqrt(a0 + a1 * ut[:, t - 1] ** 2)
    #         ut[:, t] = sigma_t[:, t] * zts[:, t]
    #     ut = np.asmatrix(ut)

    #     from varwg.time_series_analysis import time_series as ts

    #     # ts.plt.plot(zts.T)
    #     # ts.plt.plot(ut.T)
    #     # ts.plot_auto_corr(ut, k_range=7)
    #     # ts.plt.show()

    #     params = models.MGARCH_ML(ut, 2, 2)
    #     ut_sim = models.MGARCH_sim(params, T, np.cov(ut))

    #     ts.plt.plot_auto_corr([ut.T, ut_sim.T], k_range=7)
    #     ts.plt.show()

    def test_VAR_LS_sim_VAR_residuals(self):
        """Round-trip test: do we get predifined residuals back from
        VAR_LS_residuals when simulating with VAR_LS_sim?"""
        T = 5
        residuals_test = np.array(
            [
                varwg.rng.multivariate_normal(VAR_K * [0], sigma_u_test)
                for t in range(T)
            ]
        ).reshape((VAR_K, T))
        sim = models.VAR_LS_sim(
            B_test, sigma_u_test, T, u=residuals_test, n_presim_steps=0
        )
        residuals = models.VAR_residuals(sim, B_test, VAR_p)
        npt.assert_almost_equal(residuals, residuals_test)

    def test_VAR_LS_B_recover(self):
        T = 10000
        varwg.reseed(0)
        sim = models.VAR_LS_sim(B_test, sigma_u_test, T)
        # fit on simulated values check if we can recover B
        B_fit, sigma_u_fit = models.VAR_LS(sim, p=VAR_p)
        try:
            npt.assert_almost_equal(B_fit, B_test, decimal=1)
            npt.assert_almost_equal(sigma_u_fit, sigma_u_test, decimal=4)
        except AssertionError:
            import matplotlib.pyplot as plt

            fig, axs = plt.subplots(nrows=2, ncols=2, constrained_layout=True)
            axs[0, 0].matshow(B_test)
            axs[0, 0].set_title("B_test")
            axs[1, 0].matshow(B_fit)
            axs[1, 0].set_title("B_fit")
            axs[0, 1].matshow(sigma_u_test)
            axs[0, 1].set_title("sigma_u_test")
            axs[1, 1].matshow(sigma_u_fit)
            axs[1, 1].set_title("sigma_u_fit")
            plt.show()
            raise

    def test_VAREX_LS_sim_VAREX_residuals(self):
        """Round-trip test: do we get predifined residuals back from
        VAREX_LS_residuals when simulating with VAREX_LS_sim?"""
        T = 5
        residuals_test = np.array(
            [
                varwg.rng.multivariate_normal(VAR_K * [0], sigma_u_test)
                for t in range(T)
            ]
        ).reshape((VAR_K, T))
        ex = np.full(T, 0.1)
        sim, ex_out = models.VAREX_LS_sim(
            Bex_test, sigma_u_test, T, ex, u=residuals_test, n_presim_steps=0
        )
        residuals = models.VAREX_residuals(sim, ex, Bex_test, VAR_p)
        npt.assert_almost_equal(residuals, residuals_test)
        npt.assert_almost_equal(ex_out, ex)

    def test_VAREX_LS_sim_VAREX_residuals_funcy(self):
        """Round-trip test: do we get predifined residuals back from
        VAREX_LS_residuals when simulating with VAREX_LS_sim?"""
        T = 5
        residuals_test = np.array(
            [
                varwg.rng.multivariate_normal(VAR_K * [0], sigma_u_test)
                for t in range(T)
            ]
        ).reshape((VAR_K, T))
        ex_kwds = dict(fac=1.5)

        def ex(x, fac):
            return np.mean(x[:, -1]) * fac

        sim, ex_out = models.VAREX_LS_sim(
            Bex_test,
            sigma_u_test,
            T,
            ex,
            u=residuals_test,
            n_presim_steps=0,
            ex_kwds=ex_kwds,
        )
        residuals = models.VAREX_residuals(sim, ex, Bex_test, VAR_p, ex_kwds)
        npt.assert_almost_equal(residuals, residuals_test)

    def Bs_test(self, T):
        # have to have seasonal B_test and sigma_u_test
        Bs_test = np.empty(B_test.shape + (min(T, 365),))
        sin = 0.25 * np.sin(np.arange(min(T, 365)) * 2 * np.pi / 365)
        Bs_test[...] = sin[None, None, :] * np.asarray(B_test[..., None])
        Bs_test += B_test[..., None]
        # Bs_test[:, 0] = B_test[:, 0, None]
        return Bs_test

    def sigma_u_test_s(self, T):
        sigma_u_test_s = np.empty(
            (sigma_u_test.shape[0], sigma_u_test.shape[1], min(T, 365))
        )
        sin = 0.05 * np.sin(np.arange(min(T, 365)) * 2 * np.pi / 365)
        sigma_u_test_s[:] = sin[None, None, :] * np.asarray(
            sigma_u_test[..., None]
        )
        sigma_u_test_s += sigma_u_test[..., None]
        return sigma_u_test_s

    def test_sigma_u_test_s(self):
        """Is it symmetric and positive semi-definite?"""
        T = 365
        sigma_us = self.sigma_u_test_s(T)
        sym = [
            np.all((sigma_us[..., t] - sigma_us[..., t].T) == 0)
            for t in range(T)
        ]
        self.assertTrue(np.all(sym))
        for t in range(T):
            np.linalg.cholesky(sigma_us[..., t])

    def test_SVAR_LS_sim_SVAR_residuals(self):
        """Round-trip test: do we get predifined residuals back from
        SVAR_LS_residuals when simulating with SVAR_LS_sim?"""
        T = 5 * 365
        residuals_test = np.array(
            [
                varwg.rng.multivariate_normal(VAR_K * [0], sigma_u_test)
                for t in range(T)
            ]
        ).reshape((VAR_K, T))
        doys = np.arange(1, T + 1) % 365
        # residuals_test[:] = 0
        Bs_test = self.Bs_test(T)
        models.SVAR_LS_sim.clear_cache()
        sim = models.SVAR_LS_sim(
            Bs_test,
            self.sigma_u_test_s(T),
            doys,
            u=residuals_test,
            n_presim_steps=0,
        )
        residuals = models.SVAR_residuals(sim, doys, Bs_test, VAR_p)
        residuals_test = np.asarray(residuals_test)
        # the beginning time steps will not be matched, but I do not care!
        try:
            npt.assert_almost_equal(
                residuals[:, VAR_p:], residuals_test[:, VAR_p:], decimal=6
            )
        except AssertionError:
            import matplotlib.pyplot as plt

            fig, axs = plt.subplots(
                nrows=VAR_K, ncols=2, width_ratios=(0.8, 0.2), sharex="col"
            )
            for var_i, axs_k in enumerate(axs):
                axs_k[0].plot(residuals_test[var_i], label="expected")
                axs_k[0].plot(residuals[var_i], label="actual")
                axs_k[1].scatter(
                    residuals[var_i, VAR_p:], residuals_test[var_i, VAR_p:]
                )
                axs_k[1].set_aspect("equal", "box")
            axs[0, 0].legend(loc="best")
            for ax in axs.ravel():
                ax.grid(True)
            plt.show()
            raise

    def test_SVAR_LS_B_recover(self):
        T = 500 * 365
        varwg.reseed(0)
        doys = np.arange(T) % 365
        Bs_test = self.Bs_test(T)
        sigma_u_test_s = self.sigma_u_test_s(T)
        B_test_mean = Bs_test.mean(axis=-1)
        sigma_u_test_mean = sigma_u_test_s.mean(axis=-1)
        K = Bs_test.shape[0]
        sim = models.SVAR_LS_sim(Bs_test, sigma_u_test_s, doys)
        # sigma_u_test_s[:, :] = sigma_u_test_mean[..., None]
        # sim = models.SVAR_LS_sim(Bs_test, sigma_u_test_s, doys)
        sim_stat = models.VAR_LS_sim(B_test_mean, sigma_u_test_mean, T)
        means = sim.mean(axis=1)
        means_stat = sim_stat.mean(axis=1)
        try:
            npt.assert_almost_equal(means, means_stat, decimal=2)
        except AssertionError:
            import matplotlib.pyplot as plt

            fig, axs = plt.subplots(
                nrows=K, ncols=1, sharex=True, constrained_layout=True
            )
            means_th = np.array(
                [
                    np.squeeze(models.VAR_mean(Bs_test[..., t]))
                    for t in np.arange(T) % 365
                ]
            ).T
            means_stat_th = models.VAR_mean(B_test_mean)
            for var_i, ax in enumerate(axs):
                ax.plot(sim[var_i], "k", alpha=0.25)
                ax.plot(means_th[var_i], "k")
                ax.axhline(means[var_i], color="k")
                ax.plot(sim_stat[var_i], "b", alpha=0.25)
                ax.axhline(means_stat_th[var_i], color="b")
                ax.axhline(means_stat[var_i], color="b")
            plt.show()

        # std = sim.std(axis=1)
        # std_stat = sim_stat.std(axis=1)
        # npt.assert_almost_equal(std, std_stat, decimal=3)

        # import matplotlib.pyplot as plt
        # fig, axs = plt.subplots(nrows=2, ncols=K,
        #                         figsize=(9, 6))
        # ranks = (np.arange(T) - .5) / T
        # for var_i in range(K):
        #     sim_sorted = np.sort(sim[var_i])
        #     sim_stat_sorted = np.sort(sim_stat[var_i])
        #     axs[0, var_i].plot(sim_sorted, ranks,
        #                        label="seasonal")
        #     axs[0, var_i].plot(sim_stat_sorted, ranks,
        #                        label="stationary")
        #     axs[1, var_i].plot(np.sort(sim_stat[var_i]),
        #                        np.sort(sim[var_i]))
        #     smin = min(sim_sorted[0], sim_stat_sorted[0])
        #     smax = max(sim_sorted[-1], sim_stat_sorted[-1])
        #     axs[1, var_i].plot([smin, smax], [smin, smax],
        #                        "--k")
        #     axs[1, var_i].set_xlabel("stationary")
        # axs[1, 0].set_ylabel("seasonal")
        # axs[0, 0].legend(loc="best")
        # for ax in np.ravel(axs):
        #     ax.grid(True)
        # plt.show()

        # fit on simulated values check if we can recover B and
        # sigma_u
        B_fit, sigma_u_fit = models.SVAR_LS(
            sim,
            doys,
            doy_width=10,
            fft_order=3,
            p=VAR_p,
        )
        try:
            npt.assert_almost_equal(B_fit, Bs_test, decimal=1)
            npt.assert_almost_equal(sigma_u_fit, sigma_u_test_s, decimal=4)
        except AssertionError:
            import matplotlib.pyplot as plt

            fig, axs = plt.subplots(nrows=K, ncols=2, constrained_layout=True)
            J = Bs_test.shape[1]
            for k in range(K):
                for j in range(J):
                    t = (j - 1) // VAR_p
                    k_other = (j - 1) % VAR_p
                    axs[k, 0].plot(
                        Bs_test[k, j], label=f"(k={k_other}, t={t})"
                    )
                    axs[k, 0].axhline(B_test_mean[k, j])
                for j in range(K):
                    axs[k, 1].plot(
                        sigma_u_test_s[k, j], label=f"(k={k}, t={j})"
                    )
                    axs[k, 1].axhline(sigma_u_test_mean[k, j])
                for ax in axs[k]:
                    ax.set_prop_cycle(None)
                for j in range(J):
                    axs[k, 0].plot(B_fit[k, j], "--")
                for j in range(K):
                    axs[k, 1].plot(sigma_u_fit[k, j], "--")
                for ax in axs[k]:
                    ax.legend(loc="best")
                    ax.grid(True)
            axs[0, 0].set_title("B")
            axs[0, 1].set_title("sigma_u")
            plt.show()
            raise

    def test_SVAR_LS_fill(self):
        T = 3 * 365
        varwg.reseed(0)
        doys = np.arange(T) % 365
        Bs_test = self.Bs_test(T)
        sigma_u_test_s = self.sigma_u_test_s(T)
        # introduce nans
        sim = models.SVAR_LS_sim(Bs_test, sigma_u_test_s, doys)
        sim_before = np.copy(sim)
        slices = (
            (0, slice(100, 110)),
            (-1, slice(150, 160)),
            (slice(1, None), slice(200, 210)),
            (slice(None), slice(250, 260)),
        )
        for slice_ in slices:
            sim[slice_] = np.nan
        sim_filled = models.SVAR_LS_fill(Bs_test, sigma_u_test_s, doys, sim)
        assert np.all(np.isfinite(sim_filled))
        sim_filled_nans = np.copy(sim_filled)
        # did we only touch the nans?
        for slice_ in slices:
            sim_filled_nans[slice_] = np.nan
        try:
            npt.assert_almost_equal(sim, sim_filled_nans)
        except AssertionError:
            import matplotlib.pyplot as plt

            fig, axs = plt.subplots(nrows=sim.shape[0], ncols=1, sharex=True)
            for var_i, ax in enumerate(axs):
                ax.plot(sim[var_i], "-x", label="sim")
                ax.plot(sim_before[var_i], label="sim_before")
                # ax.plot(sim_filled[var_i], label="sim_filled")
            for real_i in range(100):
                sim_filled = models.SVAR_LS_fill(
                    Bs_test, sigma_u_test_s, doys, sim
                )
                for var_i, ax in enumerate(axs):
                    ax.plot(sim_filled[var_i], color="k", alpha=0.1)
            axs[0].legend(loc="best")
            plt.show()
            raise


if __name__ == "__main__":
    npt.run_module_suite()



import sys

import matplotlib.pyplot as plt
import numpy as np
import numpy.testing as npt
from scipy import integrate

import varwg
import varwg.time_series_analysis.distributions as ds
from varwg import helpers as my

distributions = {
    name: obj
    for name, obj in list(ds.__dict__.items())
    if isinstance(obj, ds.Dist)
}
if "SkewNormal" in distributions and not ds.owens:
    del distributions["SkewNormal"]

# import pandas as pd
# from varwg import config_template
# # the rainmix distribution requires data to be initialized
# sample_met_filepath = config_template.met_file
# rain_pd = pd.read_table(sample_met_filepath, sep="\t", usecols="R")
# rain_values = np.squeeze(rain_pd[:1000].as_matrix())
# rainmix_expon = ds.RainMix(ds.expon, rain_values)
# distributions["rainmix expon"] = rainmix_expon

dists_frozen = {
    name: dist(*dist._feasible_start)
    for name, dist in list(distributions.items())
}
# avoid quantiles of 0 or 1, because that causes problems with the unbounded
# distributions
quantiles = np.linspace(1e-6, 1 - 1e-6, 11)
quantiles_more = np.linspace(0.001, 0.999, 1000)
incr = 1e-6


class Test(npt.TestCase):
    def setUp(self):
        self.verbose = False
        varwg.reseed(0)

    def test_fit(self):
        """Is fit able to reproduce parameters of a ppf-generated sample?"""
        if self.verbose:
            print()
        for dist_name, dist in list(distributions.items()):
            if self.verbose:
                print("\t", dist_name)
            known_params = dist._feasible_start
            x = dist.ppf(quantiles_more, *known_params)
            if hasattr(dist, "_additional_kwds"):
                kwds = dist._additional_kwds
            else:
                kwds = {}
            fitted_params = dist.fit(x, **kwds)
            if len(kwds) > 0:
                known_params = known_params[: -len(kwds)]
            if isinstance(dist, ds.RainMix):
                try:
                    for name, known, fitted in zip(
                        dist.parameter_names, known_params, fitted_params
                    ):
                        if name in dist.supplements_names:
                            continue
                        npt.assert_almost_equal(known, fitted, decimal=1)
                except (ValueError, AssertionError):
                    if not self.verbose:
                        raise
                    import matplotlib.pyplot as plt

                    def rel_ranks(x):
                        n = len(x)
                        return (np.arange(1, n + 1) - 0.5) / n

                    plt.plot(x, rel_ranks(x), "-o", label="x")
                    plt.plot(
                        dist.sample_data,
                        rel_ranks(dist.sample_data),
                        "-x",
                        label="sample",
                    )
                    plt.legend(loc="best")
                    plt.title(dist.name)
                    plt.show()
                    print(2 * "\t", name)
                    raise
            else:
                try:
                    npt.assert_almost_equal(
                        known_params, fitted_params, decimal=1
                    )
                except AssertionError:
                    print("\rfailed")
                    raise

    def test_mean(self):
        """Is the mean-property of Frozen working?"""
        # mean should be median for symmetric distributions
        sym_names = "norm", "student_t"
        for sym_name in sym_names:
            dist = dists_frozen[sym_name]
            npt.assert_almost_equal(dist.median, dist.mean)

    def test_pdf_integral(self):
        """Is the cdf the integral of the pdf?"""
        if self.verbose:
            print()
        for dist_name, dist_frozen in dists_frozen.items():
            if isinstance(dist_frozen.dist, ds.RainMix):
                # see specific test below
                continue
            if self.verbose:
                print("\t", dist_name)
            # get two feasible input values
            q0, q1 = 0.25, 0.99
            x0, x1 = dist_frozen.ppf([q0, q1])
            integral = integrate.quad(dist_frozen.pdf, x0, x1)[0]
            try:
                npt.assert_almost_equal(integral, q1 - q0, decimal=2)
            except AssertionError as exc:
                if not self.verbose:
                    raise
                dist_frozen.plot_fit(dist_frozen.sample(500))
                plt.show()
                print(exc)
                raise

    def test_rainmix_pdf(self):
        if self.verbose:
            print()
        rainmix_dists = {
            name: dist
            for name, dist in dists_frozen.items()
            if isinstance(dist.dist, ds.RainMix)
        }
        for dist_name, dist_frozen in rainmix_dists.items():
            if self.verbose:
                print("\t", dist_name)
            dist_frozen.dist.debug = True
            # get two feasible input values
            # q0, q1 = .01, .99
            q0, q1 = 1e-6, 1 - 1e-6
            xx = dist_frozen.ppf(np.linspace(q0, q1, 500))
            x0, x1 = xx[[0, -1]]
            f_thresh = dist_frozen.parameter_dict["f_thresh"]
            q_thresh = dist_frozen.dist.q_thresh
            # npt.assert_almost_equal(dist_frozen.ppf(q_thresh), f_thresh,
            #                         decimal=3)
            # npt.assert_almost_equal(dist_frozen.cdf(f_thresh), q_thresh,
            #                         decimal=4)
            integral_par = integrate.quad(
                dist_frozen.pdf, x0, f_thresh - 1e-9
            )[0]
            integral_kde = integrate.quad(dist_frozen.pdf, f_thresh, x1)[0]
            try:
                npt.assert_almost_equal(integral_par, q_thresh - q0, decimal=2)
            except AssertionError as exc:
                print("parametric part")
                print(exc)
                raise
            try:
                npt.assert_almost_equal(integral_kde, q1 - q_thresh, decimal=2)
            except AssertionError as exc:
                print("KDE part")
                print(exc)
                print(
                    f"exp/act: {integral_kde / (q1 - q_thresh)} "
                    f"act/exp: {(q1 - q_thresh) / integral_kde}"
                )
                raise
            # integral = integrate.quad(dist_frozen.pdf, x0, x1)[0]
            # try:
            #     npt.assert_almost_equal(integral, q1 - q0, decimal=2)
            # except AssertionError as exc:
            #     # if not self.verbose:
            #     #     raise
            #     print("Full distribution")
            #     print(exc)
            #     sample = dist_frozen.sample(5000)
            #     fig, (ax1, ax2) = dist_frozen.plot_fit(sample)
            #     ax2.axhline(dist_frozen.dist.q_thresh, linestyle="--")
            #     ax2.axvline(
            #         dist_frozen.parameter_dict["f_thresh"], linestyle="--"
            #     )
            #     # pdf by cdf differentiation
            #     cdf_diff = np.array(
            #         [derivative(dist_frozen.cdf, x0, dx=1e-6) for x0 in xx]
            #     )
            #     ax2.plot(xx, cdf_diff, "y--")
            #     ax2.set_ylim([0, 1])
            #     plt.show()
            #     raise

    def test_roundtrip(self):
        """Is the ppf really the inverse of the cdf?"""
        if self.verbose:
            print()
        for dist_name, dist_frozen in dists_frozen.items():
            dist_frozen.dist.debug = True
            if self.verbose:
                print("\t", dist_name)
            x = dist_frozen.ppf(quantiles_more)
            quantiles_back = dist_frozen.cdf(x)
            try:
                # distributions with lower thresholds will fail to
                # reproduce quantiles below it
                rain_prob = dist_frozen.parameter_dict["rain_prob"]
                dry_mask = quantiles_more < (1 - rain_prob)
                quantiles_back[dry_mask] = quantiles_more[dry_mask]
            except KeyError:
                pass
            npt.assert_array_less(quantiles_back, 1.0)
            npt.assert_array_less(0, quantiles_back + sys.float_info.min)
            try:
                # if isinstance(dist_frozen.dist, ds.RainMix):
                #     quantiles_back -= 1e-4
                npt.assert_almost_equal(
                    quantiles_more, quantiles_back, decimal=2
                )
            except AssertionError:
                if not self.verbose:
                    raise
                import matplotlib.pyplot as plt

                def rel_ranks(x):
                    n = len(x)
                    return (np.arange(n) - 0.5) / n

                fig, axs = plt.subplots(
                    ncols=2,
                    sharey=True,
                    figsize=(8, 4),
                    # subplot_kw=dict(aspect="equal")
                )
                axs[0].scatter(quantiles_more, quantiles_back, marker="x")
                if isinstance(dist_frozen.dist, ds.RainMix):
                    dist = dist_frozen.dist
                    axs[1].plot(x, rel_ranks(x), "-x", label="x")
                    axs[1].plot(
                        dist.sample_data,
                        rel_ranks(dist.sample_data),
                        "-x",
                        label="sample_data",
                    )
                    axs[1].legend(loc="best")
                    axs[0].axvline(dist.q_thresh)
                    axs[0].axhline(dist.q_thresh)
                    axs[1].axhline(dist.q_thresh)
                    axs[1].axvline(dist_frozen.parameter_dict["f_thresh"])
                    # sample_quantiles = dist_frozen.cdf(dist.sample_data)
                    # ax.scatter(quantiles_more, sample_quantiles,
                    #            marker="o",
                    #            # facecolor=4 * (0,),
                    #            # edgecolor=4 * (1,),
                    # )
                axs[0].plot([0, 1], [0, 1])
                axs[0].set_xlabel("quantiles_more")
                axs[0].set_ylabel("quantiles_back")
                axs[0].grid(True)
                axs[1].grid(True)
                fig.suptitle(dist_frozen.name)
                plt.show()
                raise

    def test_sample(self):
        if self.verbose:
            print()
        for dist_name, dist_frozen in list(dists_frozen.items()):
            if self.verbose:
                print("\t", dist_name)
            sample_size = 100
            sample = dist_frozen.sample(sample_size)
            npt.assert_equal(sample.size, sample_size)

    def test_scalar(self):
        """Do pdf, cdf and ppf accept and return scalar values?"""
        if self.verbose:
            print()
        for dist_name, dist_frozen in list(dists_frozen.items()):
            if isinstance(dist_name, ds._KDE):
                continue
            if self.verbose:
                print("\t", dist_name)
            self.assertTrue(np.isscalar(dist_frozen.pdf(1)))
            self.assertTrue(np.isscalar(dist_frozen.cdf(1)))
            self.assertTrue(np.isscalar(dist_frozen.ppf(0.5)))

    def test_out_of_bounds_pdf(self):
        """Do pdfs return nan for input outside the range?"""
        if self.verbose:
            print()
        for dist_name, dist_frozen in list(dists_frozen.items()):
            if isinstance(dist_frozen.dist, ds.RainMix):
                continue
            pars = dist_frozen.parameter_dict
            lower = pars.get("lc") or pars.get("l")
            upper = pars.get("uc") or pars.get("u")
            if lower:
                if self.verbose:
                    print("\t lower bound of ", dist_name)
                ret = dist_frozen.pdf(lower - incr)
                self.assertTrue(np.isnan(ret), ret)
            if upper:
                if self.verbose:
                    print("\t upper bound of ", dist_name)
                ret = dist_frozen.pdf(upper + incr)
                self.assertTrue(np.isnan(ret), ret)

    # def test_out_of_bounds_cdf(self):
    #     """Do cdfs return nan for input outside the range?"""
    #     if self.verbose:
    #         print()
    #     for dist_name, dist_frozen in list(dists_frozen.items()):
    #         if isinstance(dist_frozen.dist, ds.RainMix):
    #             continue
    #         pars = dist_frozen.parameter_dict
    #         lower = pars.get("lc") or pars.get("l")
    #         upper = pars.get("uc") or pars.get("u")
    #         if lower:
    #             if self.verbose:
    #                 print("\t lower bound of ", dist_name)
    #             ret = dist_frozen.cdf(lower - incr)
    #             self.assertTrue(np.isnan(ret), ret)
    #         if upper:
    #             if self.verbose:
    #                 print("\t upper bound of ", dist_name)
    #             ret = dist_frozen.cdf(upper + incr)
    #             self.assertTrue(np.isnan(ret), ret)

    # def test_out_of_bounds_ppf(self):
    #     """Do ppfs return nan for input outside the range?"""
    #     if self.verbose:
    #         print()
    #     for dist_name, dist_frozen in list(dists_frozen.items()):
    #         if isinstance(dist_frozen.dist, ds.RainMix):
    #             continue
    #         pars = dist_frozen.parameter_dict
    #         lower = pars.get("lc") or pars.get("l")
    #         upper = pars.get("uc") or pars.get("u")
    #         if lower:
    #             if self.verbose:
    #                 print("\t lower bound of ", dist_name)
    #             ret = dist_frozen.ppf(lower - incr)
    #             self.assertTrue(np.isnan(ret), ret)
    #         if upper:
    #             if self.verbose:
    #                 print("\t upper bound of ", dist_name)
    #             ret = dist_frozen.ppf(upper + incr)
    #             self.assertTrue(np.isnan(ret), ret)

    def test_rain(self):
        thresh = 0.1
        exp = ds.expon(0.0, 1.0)
        exp_rv = exp.ppf(quantiles_more)
        # rain and no rain in it
        exp_rv = np.concatenate((np.zeros(len(quantiles_more) // 2), exp_rv))
        rain_dist = ds.Rain(ds.expon, thresh)
        rain = rain_dist(*rain_dist.fit(exp_rv))
        ranks = my.rel_ranks(exp_rv)
        rain_qq = rain.cdf(exp_rv)
        npt.assert_almost_equal(
            ranks[exp_rv > thresh], rain_qq[exp_rv > thresh], decimal=3
        )
        rain_retrans = rain.ppf(rain_qq)
        npt.assert_almost_equal(
            exp_rv[rain_retrans > thresh], rain_retrans[rain_retrans > thresh]
        )


if __name__ == "__main__":
    import warnings

    warnings.simplefilter("error", RuntimeWarning)
    npt.run_module_suite()



"""Download DWD test fixtures for offline testing."""

from pathlib import Path
import xarray as xr
import dwd_opendata

FIXTURE_DIR = Path(__file__).parent
FIXTURE_DIR.mkdir(exist_ok=True)


def download_stoetten_precipitation():
    """Download Stötten precipitation data for test_brunner and test_meteox2y."""
    print("Downloading Stötten precipitation...")
    prec = dwd_opendata.load_station("Stötten", "precipitation", time="hourly")
    prec = (
        prec.interpolate_na("time")
        .squeeze()
        .resample(time="1d")
        .sum()
        .sel(time=slice("1990", "2020"))
        .dropna("time")
    )
    prec.to_netcdf(FIXTURE_DIR / "stoetten_precipitation_daily_1990_2020.nc")
    print(
        f"Saved to {FIXTURE_DIR / 'stoetten_precipitation_daily_1990_2020.nc'}"
    )


def download_konstanz_temperature():
    """Download Konstanz air temperature for test_seasonal_distributions."""
    print("Downloading Konstanz air temperature...")
    theta_xr = dwd_opendata.load_station(
        "Konstanz", "air_temperature"
    ).squeeze()
    theta_xr = (
        theta_xr.sel(time=slice("2000", "2016"))
        .resample(time="D")
        .mean()
        .interpolate_na("time")
    )
    theta_xr.to_netcdf(FIXTURE_DIR / "konstanz_temperature_daily_2000_2016.nc")
    print(
        f"Saved to {FIXTURE_DIR / 'konstanz_temperature_daily_2000_2016.nc'}"
    )


def download_konstanz_sun():
    """Download Konstanz sun data for test_seasonal_distributions."""
    print("Downloading Konstanz sun...")
    sun_xr = dwd_opendata.load_station("Konstanz", "sun").squeeze()
    sun_xr = (
        sun_xr.sel(time=slice("2000", "2016"))
        .resample(time="D")
        .mean()
        .interpolate_na("time")
    )
    sun_xr.to_netcdf(FIXTURE_DIR / "konstanz_sun_daily_2000_2016.nc")
    print(f"Saved to {FIXTURE_DIR / 'konstanz_sun_daily_2000_2016.nc'}")


def download_freiburg_precipitation():
    """Download Freiburg precipitation for test_seasonal_distributions."""
    print("Downloading Freiburg precipitation...")
    prec_xr = dwd_opendata.load_station("Freiburg", "precipitation")
    prec_xr = (
        prec_xr.sel(time=slice("2000", "2016"))
        .resample(time="D")
        .sum()
        .interpolate_na("time")
    )
    prec_xr.to_netcdf(
        FIXTURE_DIR / "freiburg_precipitation_daily_2000_2016.nc"
    )
    print(
        f"Saved to {FIXTURE_DIR / 'freiburg_precipitation_daily_2000_2016.nc'}"
    )


if __name__ == "__main__":
    download_stoetten_precipitation()
    download_konstanz_temperature()
    download_konstanz_sun()
    download_freiburg_precipitation()
    print("\nAll fixtures downloaded successfully!")



"""This provides common ground for seasonal_distributions and seasonal_kde."""

import calendar
import hashlib
import numpy as np
import scipy.optimize as sp_optimize
from varwg import times
from varwg import helpers as my
from varwg.smoothing import smooth
from varwg.time_series_analysis import phase_randomization, distributions


def build_doy_mask(doys, doy_width, doys_unique=None):
    """Mask to access doy neighborhood in data."""
    if doys_unique is None:
        doys_unique = np.unique(doys)
    doy_mask = np.empty((len(doys_unique), len(doys)), dtype=bool)
    for doy_i, doy in enumerate(doys_unique):
        ii = (doys > doy - doy_width) & (doys <= doy + doy_width)
        if (doy - doy_width) < 0:
            ii |= doys > (365.0 - doy_width + doy)
        if (doy + doy_width) > 365:
            ii |= doys < (doy + doy_width - 365.0)
        doy_mask[doy_i] = ii
    # # take into account whether doys do not start at 1
    # doy_mask = np.roll(doy_mask, -doys.argmin())
    return doy_mask


class Seasonal(object):

    def __init__(self, data, datetimes, kill_leap=False):
        finite_mask = np.isfinite(data)
        self.data = data[finite_mask]
        self.datetimes = datetimes[finite_mask]
        # self.timestep = ((self.datetimes[1] -
        #                   self.datetimes[0]).total_seconds() //
        #                  (60 ** 2 * 24))
        self.timestep = (
            self.datetimes[1] - self.datetimes[0]
        ).total_seconds() / (60**2 * 24)
        assert self.timestep > 0
        self.doys = times.datetime2doy(datetimes[finite_mask])
        self.doys_unique = np.unique(
            my.round_to_float(self.doys, self.timestep)
        )
        # we could have timestamps like "2004-01-01T00:30:00", so we
        # might need an offset for the unique doys
        doy0_diff = self.doys.min() - self.doys_unique[0]
        self.doys_unique += doy0_diff
        assert np.max(self.doys_unique) <= 367
        self.dt = self.doys_unique[1] - self.doys_unique[0]
        self.n_doys = len(self.doys_unique)
        if kill_leap:
            # get rid off feb 29
            self._feb29()
        self._qq_shift_cache = None

    def _feb29(self):
        # get rid off feb 29
        feb29_mask = ~times.feb29_mask(self.datetimes)
        self.datetimes = self.datetimes[feb29_mask]
        self.doys = times.datetime2doy(self.datetimes)
        self.data = self.data[feb29_mask]
        # doys should be from 0-365
        isleap = np.array([calendar.isleap(dt.year) for dt in self.datetimes])
        self.doys[isleap & (self.doys >= 31 + 29)] -= 1
        self.doys_unique = np.unique(
            my.round_to_float(self.doys, self.timestep)
        )
        self.n_doys = len(self.doys_unique)

    def duplicate_feb28(self, doys):
        """Put doys in the range [1, 365] by duplicating February 28th and
        decrementing all following doys of the year.
        """
        doys_new = []
        doys_splitted = np.split(doys, np.where(doys.astype(int) == 1)[0])
        # the first element is empty when we split on doy==1 if first
        # element of doys is 1!
        if len(doys_splitted[0]) == 0:
            doys_splitted = doys_splitted[1:]
        for doys_one_year in doys_splitted:
            if np.isclose(doys_one_year[-1], 366):
                doys_one_year[doys_one_year >= 29] -= 1
            doys_new += [doys_one_year]
        doys_new = np.concatenate(doys_new)
        assert len(doys_new) == len(doys)
        return doys_new

    def doys2doys_ii(self, doys):
        """Doys indices from doy values."""
        doys = np.atleast_1d(doys)
        # use the parameters of 28. feb for 29.feb
        year_end_ii = np.where(np.diff(doys) < 0)[0] + 1
        year_end_ii = np.concatenate(([0], year_end_ii, [len(doys)]))
        for start_i, end_i in zip(year_end_ii[:-1], year_end_ii[1:]):
            year_slice = slice(start_i, end_i)
            year_doys = doys[year_slice]
            if np.max(year_doys) >= 366:
                year_doys[year_doys > 31 + 29] -= 1
                doys[year_slice] = year_doys
        doys_ii = ((doys - 1) / self.dt).astype(int)
        if len(doys_ii) < len(doys):
            doys_ii = [my.val2ind(self.doys_unique, doy) for doy in doys]
        return doys_ii

    def qq_shift(self, theta_incr, trig_pars, x=None, doys=None, **kwds):
        """Empirical estimation of shift in std-normal for given theta_incr."""
        # executing this method is expensive and in the context of
        # weathercop's conditional simulation, might happen often with
        # the same theta_incr, so do some caching here to return known
        # results
        if self._qq_shift_cache is None:
            self._qq_shift_cache = {}
        # theta_incr_key = hash(round(theta_incr[0], 6))
        theta_incr_key = hashlib.md5(
            str(round(theta_incr[0], 6)).encode()
        ).hexdigest()
        if theta_incr_key not in self._qq_shift_cache:
            qq = self.cdf(trig_pars, x=x, doys=doys, **kwds)
            zero = 1e-6
            one = 1 - zero
            stdn = distributions.norm.ppf(
                np.minimum(np.maximum(qq, zero), one)
            )
            stdn = phase_randomization.randomize2d(np.array([stdn])).squeeze()
            data = np.atleast_1d(self.data if x is None else x)
            data_mean = data.mean()

            def c2incr(c):
                xx_act = self.ppf(
                    trig_pars,
                    quantiles=distributions.norm.cdf(stdn + c),
                    doys=doys,
                    kwds=kwds,
                )
                return (xx_act.mean() - data_mean - theta_incr) ** 2

            # if self.verbose:
            #     print(f"\tFilling shift-cache for {theta_incr=}")
            result = sp_optimize.minimize_scalar(c2incr)

            # fig, ax = plt.subplots(nrows=1, ncols=1)
            # xx_act = self.ppf(trig_pars,
            #                   quantiles=distributions.norm.cdf(stdn + result.x),
            #                   doys=doys,
            #                   kwds=kwds)
            # ax.axvline(data_mean, label="data_mean", color="k")
            # ax.axvline(xx_act.mean(), label="xx_act", color="b")
            # ax.axvline(data_mean + theta_incr, label="data_mean + theta_incr",
            #            color="k", linestyle="--")
            # ax.legend()
            # plt.show()

            self._qq_shift_cache[theta_incr_key] = result.x
        return self._qq_shift_cache[theta_incr_key]

    @property
    def n_years(self):
        return self.datetimes[-1].year - self.datetimes[0].year + 1

    @property
    def hours(self):
        # general reminder: self.doys are floats, they are just a
        # representation of the date with all information except the year
        # self.hours are just the hours, and no information about minutes or
        # seconds are given
        return times.datetime2hour(self.datetimes)

    @property
    def hours_per_day(self):
        return int(self.timestep**-1)

    def _set_monthly_ticks(self, ax):
        ax.set_xticks((1, 32, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335))
        ax.set_xticklabels(
            (
                "Jan",
                "Feb",
                "Mar",
                "Apr",
                "May",
                "Jun",
                "Jul",
                "Aug",
                "Sep",
                "Oct",
                "Nov",
                "Dec",
            ),
            rotation=45,
        )

    def rain_probs(self, threshold, doys=None):
        if doys is None:
            doys = self.doys
        probs_per_doy = np.zeros_like(self.doys_unique)
        for doy_i in self.doys_unique.astype(int) - 1:
            data = self.data[self.doy_mask[doy_i]]
            probs_per_doy[doy_i] = np.mean(data > threshold)
        probs_per_doy = smooth(probs_per_doy, self.doy_width, periodic=True)
        return probs_per_doy[doys.astype(int) - 1]


class Torus(Seasonal):

    def __init__(self, hour_neighbors):
        self.hour_neighbors = hour_neighbors
        # for property caching
        self._torus = None

    @property
    def torus(self):
        if self._torus is None:
            self._torus = self._construct_torus(self.data)
        return self._torus

    def _construct_torus(self, values, hours=None, doys=None, years=None):
        """Returns a 3d array representation of the 1d input.

        Parameters
        ----------
        values : 1d array
        """
        # hours, doys and years are constructed so that they can be
        # used as indices for the torus
        if hours is None:
            hours = self.hours
        if doys is None:
            doys = self.doys
        if years is None:
            first_year = self.datetimes[0].year
            years = [dt.year - first_year for dt in self.datetimes]
            n_years = self.n_years
        else:
            n_years = len(np.unique(years))
        hours = list(hours.astype(int))
        doys = list(doys.astype(int) - 1)
        years = list(years)

        # we deleted the 29 of february to have a sane and full array
        # with 365 days in the day dimension
        torus = np.full((self.hours_per_day, 365, n_years), np.nan)
        torus[[hours, doys, years]] = values
        torus = self._pad_torus(torus)

        # import matplotlib.pyplot as plt
        # fig, axs = plt.subplots(len(np.unique(years)))
        # for year_i, ax in enumerate(np.ravel(axs)):
        #     # , cmap=plt.get_cmap("BuPu"))
        #     cm = ax.matshow(torus[..., year_i])
        # # plt.colorbar(cm)
        # plt.show()
        return torus

    def torus_fft(
        self,
        values,
        hours=None,
        doys=None,
        years=None,
        fft_order=5,
        padded=False,
    ):
        if np.ndim(values) == 1:
            values = self._construct_torus(values, hours, doys, years)
        if not padded:
            values = self._pad_torus(values)
        omega = np.fft.fft2(values)
        # # avoid long hourly frequencies
        # omega[24:-24] = 0
        if fft_order is not None:
            nth_largest = np.sort(np.ravel(np.abs(omega)))[-fft_order]
            omega[np.abs(omega) < nth_largest] = 0
        return np.squeeze(omega)

    def smooth_torus(self, values, fft_order=5, padded=False):
        omega = self.torus_fft(values, fft_order=5, padded=padded)
        return np.fft.irfft2(omega, s=values.shape)

    def _pad_torus(self, torus):
        # to achieve periodicity, i.e. move up -> advance in hours,
        # move right -> advance in days
        torus = np.vstack(
            (
                np.roll(torus[-self.hour_neighbors :], -1, axis=1),
                torus,
                np.roll(torus[: self.hour_neighbors], 1, axis=1),
            )
        )
        torus = np.hstack(
            (torus[:, -self.doy_width :], torus, torus[:, : self.doy_width])
        )
        return torus

    def _unpad_torus(self, torus):
        return torus[
            self.hour_neighbors : -self.hour_neighbors,
            self.doy_width : -self.doy_width,
        ]

    def _unpadded_index(self, doy):
        doy = np.atleast_1d(doy)
        hour_index = ((doy - doy.astype(int)) * self.hours_per_day).astype(int)
        doy_index = doy.astype(int) - 1
        return np.squeeze(hour_index), np.squeeze(doy_index)

    def _torus_index(self, doy):
        """Returns the index corresponding to a decimal doy."""
        hour_index, doy_index = self._unpadded_index(doy)
        return (hour_index + self.hour_neighbors, doy_index + self.doy_width)

    def _torus_slice(self, doy):
        """Returns slice of the torus centered around doy."""
        hour_index, doy_index = self._torus_index(doy)
        hour_slice = slice(
            hour_index - self.hour_neighbors,
            hour_index + self.hour_neighbors + 1,
        )
        doy_slice = slice(
            doy_index - self.doy_width, doy_index + self.doy_width + 1
        )
        return hour_slice, doy_slice

    @property
    def doy_hour_weights(self):
        """To be used as a kernel to weight distances in the doy-hour domain."""
        if self._doy_hour_weights is None:
            hour_slice, doy_slice = self._torus_slice(0)
            n_hours = hour_slice.stop - hour_slice.start
            n_doys = doy_slice.stop - doy_slice.start
            # distance in the two temporal dimensions
            hour_dist, doy_dist = np.meshgrid(
                list(range(n_doys)), list(range(n_hours))
            )
            hour_middle = n_hours // 2
            doy_middle = n_doys // 2
            time_distances = np.empty((n_hours, n_doys, self.n_years))
            temp = np.sqrt(
                (hour_dist - hour_middle) ** 2 + (doy_dist - doy_middle) ** 2
            )
            time_distances[:] = temp[..., None]
            self._doy_hour_weights = time_distances
        return self._doy_hour_weights.ravel()
